# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['newline_and_indent']

from . import pyparse
from .common import get_text

def newline_and_indent(sourceview, INDENT_WIDTH):
    """
    Get a sourceview. Add a newline and indent - what happens when the user
    pressed Enter.
    """
    # This is based on newline_and_indent_event(),
    # from idlelib/EditorWindow.py
    sb = sourceview.get_buffer()
    sb.begin_user_action()
    insert_mark = sb.get_insert()
    insert = lambda: sb.get_iter_at_mark(insert_mark)
    try:
        sb.delete_selection(True, True)
        line = get_text(sb, sb.get_iter_at_line(insert().get_line()), insert())
        i, n = 0, len(line)
        while i < n and line[i] in " \t":
            i = i+1
        if i == n:
            # the cursor is in or at leading indentation in a continuation
            # line; just copy the indentation
            sb.insert_at_cursor('\n'+line)
            sourceview.scroll_mark_onscreen(sb.get_insert())
            return True
        indent = line[:i]
        # strip whitespace before insert point
        i = 0
        while line and line[-1] in " \t":
            line = line[:-1]
            i = i+1
        if i:
            sb.delete(sb.get_iter_at_line_offset(insert().get_line(),
                                                 len(line)),
                      insert())
        # strip whitespace after insert point
        it = insert(); it.forward_to_line_end()
        after_insert = get_text(sb, insert(), it)
        i = 0
        while i < len(after_insert) and after_insert[i] in " \t":
            i += 1
        if i > 0:
            it = insert(); it.forward_chars(i)
            sb.delete(insert(), it)
        # start new line
        sb.insert_at_cursor('\n')
        # scroll to see the beginning of the line
        sourceview.scroll_mark_onscreen(sb.get_insert())
        #self.scrolledwindow_sourceview.get_hadjustment().set_value(0)

        # adjust indentation for continuations and block
        # open/close first need to find the last stmt
        y = pyparse.Parser(INDENT_WIDTH, INDENT_WIDTH)
        y.set_str(get_text(sb, sb.get_start_iter(), insert()))
        c = y.get_continuation_type()
        if c != pyparse.C_NONE:
            # The current stmt hasn't ended yet.
            if c == pyparse.C_STRING_FIRST_LINE:
                # after the first line of a string; do not indent at all
                pass
            elif c == pyparse.C_STRING_NEXT_LINES:
                # inside a string which started before this line;
                # just mimic the current indent
                sb.insert_at_cursor(indent)
            elif c == pyparse.C_BRACKET:
                # line up with the first (if any) element of the
                # last open bracket structure; else indent one
                # level beyond the indent of the line with the
                # last open bracket
                sb.insert_at_cursor(' ' * y.compute_bracket_indent())
            elif c == pyparse.C_BACKSLASH:
                # if more than one line in this stmt already, just
                # mimic the current indent; else if initial line
                # has a start on an assignment stmt, indent to
                # beyond leftmost =; else to beyond first chunk of
                # non-whitespace on initial line
                if y.get_num_lines_in_stmt() > 1:
                    sb.insert_at_cursor(indent)
                else:
                    sb.insert_at_cursor(' ' * y.compute_backslash_indent())
            else:
                assert False, "bogus continuation type %r" % (c,)
            return True

        # This line starts a brand new stmt; indent relative to
        # indentation of initial line of closest preceding
        # interesting stmt.
        indent = len(y.get_base_indent_string())
        if y.is_block_opener():
            indent = (indent // INDENT_WIDTH + 1) * INDENT_WIDTH
        elif y.is_block_closer():
            indent = max(((indent - 1) // INDENT_WIDTH) * INDENT_WIDTH, 0)
        sb.insert_at_cursor(' ' * indent)
        return True
    finally:
        sb.end_user_action()




#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

all = ['update_check']

import threading
import httplib
import json

try:
    from glib import idle_add
except ImportError:
    # In PyGObject 2.14, it's in gobject.
    from gobject import idle_add

from .. import release_timestamp
from .git import get_commit_details
from . import bug_report

def log(s):
    pass

def update_check_in_thread(is_git, cur_time, on_update_available):
    if is_git:
        fn = '/latest-commit.json'
    else:
        fn = '/latest-release.json'

    try:
        conn = httplib.HTTPConnection("www.dreampie.org")
        log("Fetching http://www.dreampie.org%s" % fn)
        conn.request("GET", fn)
        r = conn.getresponse()
        if r.status != 200:
            return
        data = r.read()
        d = json.loads(data)
    except Exception, e:
        log("Exception while fetching update info: %s" % e)
        return
    
    if is_git:
        latest_name = None
        latest_time = d['latest_commit_timestamp']
    else:
        latest_name = d['latest_release_name']
        latest_time = d['latest_release_timestamp']
    
    bug_report.set_update_info(is_git, latest_name, latest_time, cur_time)
    if latest_time > cur_time:
        # Call in main thread
        idle_add(on_update_available, is_git, latest_name, latest_time)
    else:
        log("No more recent release/commit")

def update_check(on_update_available):
    """
    Check (in the background) if updates are available.
    If so, on_update_available(is_git, latest_name, latest_time) will be called.
    """
    commit_id, commit_time = get_commit_details()
    if commit_id is not None:
        is_git = True
        cur_time = commit_time
    else:
        is_git = False
        cur_time = release_timestamp
    
    t = threading.Thread(target=update_check_in_thread,
                         args=(is_git, cur_time, on_update_available))
    t.daemon = True
    t.start()
    


# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['StatusBar']

try:
    from glib import timeout_add_seconds, source_remove
except ImportError:
    timeout_add_seconds = None
    # In PyGObject 2.14, it's in gobject.
    from gobject import timeout_add, source_remove

class StatusBar(object):
    """
    Add messages to the status bar which disappear when the contents is changed.
    """
    def __init__(self, sourcebuffer, sv_changed, statusbar):
        self.sourcebuffer = sourcebuffer
        sv_changed.append(self.on_sv_changed)
        self.statusbar = statusbar
        
        # id of a message displayed in the status bar to be removed when
        # the contents of the source buffer is changed
        self.sourcebuffer_status_id = None
        self.sourcebuffer_changed_handler_id = None
        
        self.timeout_handle = None

    def on_sv_changed(self, new_sv):
        if self.sourcebuffer_status_id is not None:
            self.clear_status()
        self.sourcebuffer = new_sv.get_buffer()
    
    def set_status(self, message):
        """Set a message in the status bar to be removed when the contents
        of the source buffer is changed"""
        if self.sourcebuffer_status_id is not None:
            self.clear_status()
        self.sourcebuffer_status_id = self.statusbar.push(0, message)
        self.sourcebuffer_changed_handler_id = \
            self.sourcebuffer.connect('changed', self.on_sourcebuffer_changed)
        
        if timeout_add_seconds is not None:
            timeout_add_seconds(10, self.on_timeout)
        else:
            timeout_add(10000, self.on_timeout)

    def clear_status(self):
        try:
            self.statusbar.remove_message(0, self.sourcebuffer_status_id)
        except AttributeError:
            # Support older PyGTK
            self.statusbar.remove(0, self.sourcebuffer_status_id)
        self.sourcebuffer_status_id = None
        self.sourcebuffer.disconnect(self.sourcebuffer_changed_handler_id)
        self.sourcebuffer_changed_handler_id = None
        
        if self.timeout_handle is not None:
            source_remove(self.timeout_handle)
            self.timeout_handle = None
    
    def on_sourcebuffer_changed(self, _widget):
        self.clear_status()
        return False
    
    def on_timeout(self):
        if self.sourcebuffer_status_id is not None:
            self.clear_status()
        return False
    
    



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['write_command']
import tokenize
import keyword

from .tags import PROMPT, COMMAND, COMMAND_DEFS, COMMAND_SEP

from .tags import KEYWORD, BUILTIN, STRING, NUMBER, COMMENT

keywords = set(keyword.kwlist)
builtins = set(__builtins__)

def write_command(write, command):
    """Write a command to the textview, with syntax highlighting and "...".
    """
    lines = [x+'\n' for x in command.split('\n')]
    # Remove last newline - we don't tag it with COMMAND to separate commands
    lines[-1] = lines[-1][:-1]
    defs_lines = get_defs_lines(lines)
    tok_iter = tokenize.generate_tokens(iter(lines).next)
    highs = []
    for typ, token, (sline, scol), (eline, ecol), line in tok_iter:
        tag = None
        if typ == tokenize.NAME:
            if token in keywords:
                tag = KEYWORD
            elif token in builtins:
                tag = BUILTIN
        elif typ == tokenize.STRING:
            tag = STRING
        elif typ == tokenize.NUMBER:
            tag = NUMBER
        elif typ == tokenize.COMMENT:
            tag = COMMENT
        if tag is not None:
            highs.append((tag, sline-1, scol, eline-1, ecol))
    # Adding a terminal highlight will help us avoid end-cases
    highs.append((None, len(lines), 0, len(lines), 0))

    def my_write(s, is_defs, *tags):
        if not is_defs:
            write(s, *tags)
        else:
            write(s, COMMAND_DEFS, *tags)

    high_pos = 0
    cur_high = highs[0]
    in_high = False
    for lineno, line in enumerate(lines):
        is_defs = defs_lines[lineno]
            
        if lineno != 0:
            my_write('... ', is_defs, COMMAND, PROMPT)
        col = 0
        while col < len(line):
            if not in_high:
                if cur_high[1] == lineno:
                    if cur_high[2] > col:
                        my_write(line[col:cur_high[2]], is_defs, COMMAND)
                        col = cur_high[2]
                    in_high = True
                else:
                    my_write(line[col:], is_defs, COMMAND)
                    col = len(line)
            else:
                if cur_high[3] == lineno:
                    if cur_high[4] > col:
                        my_write(line[col:cur_high[4]],
                                 is_defs, COMMAND, cur_high[0])
                        col = cur_high[4]
                    in_high = False
                    high_pos += 1
                    cur_high = highs[high_pos]
                else:
                    my_write(line[col:], is_defs, COMMAND, cur_high[0])
                    col = len(line)
    write('\n', COMMAND)
    write('\r', COMMAND_SEP)

def get_defs_lines(lines):
    """
    Get a list of lines - strings with Python code.
    Return a list of booleans - whether a line should be hidden when hide-defs
    is True, because it's a part of a function or class definitions.
    """
    # return value
    defs_lines = [False for _line in lines]
    # Last line with a 'def' or 'class' NAME
    last_def_line = -2
    # Indentation depth - when reaches 0, we are back in a non-filtered area.
    cur_depth = 0
    # First line of current filtered area
    first_filtered_line = None
    
    tok_iter = tokenize.generate_tokens(iter(lines).next)
    for typ, token, (sline, _scol), (_eline, _ecol), _line in tok_iter:
        if cur_depth > 0:
            if typ == tokenize.INDENT:
                cur_depth += 1
            elif typ == tokenize.DEDENT:
                cur_depth -= 1
                if cur_depth == 0:
                    for i in range(first_filtered_line, sline-1):
                        defs_lines[i] = True
                    first_filtered_line = None
        else:
            if typ == tokenize.NAME and token in ('def', 'class'):
                last_def_line = sline
            elif typ == tokenize.INDENT and sline == last_def_line + 1:
                cur_depth = 1
                first_filtered_line = sline-1
    
    return defs_lines



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['CallTips']

import gtk

try:
    from glib import idle_add
except ImportError:
    from gobject import idle_add

from .hyper_parser import HyperParser
from .call_tip_window import CallTipWindow
from .common import beep, get_text

class CallTips(object):
    def __init__(self, sourceview, sv_changed, window_main, get_func_doc,
                 INDENT_WIDTH):
        self.sourceview = sourceview
        self.sourcebuffer = sb = sourceview.get_buffer()
        sv_changed.append(self.on_sv_changed)
        self.window_main = window_main
        self.get_func_doc = get_func_doc
        self.INDENT_WIDTH = INDENT_WIDTH

        self.ctwindow = CallTipWindow(sourceview, sv_changed)

        self.start_mark = sb.create_mark(None, sb.get_start_iter(),
                                         left_gravity=True)
        self.end_mark = sb.create_mark(None, sb.get_start_iter(),
                                       left_gravity=False)

        self.is_shown = False

        # A list with (widget, handler) pairs, to be filled with self.connect()
        self.signals = []

    def on_sv_changed(self, new_sv):
        sb = self.sourcebuffer
        self.hide()
        sb.delete_mark(self.start_mark)
        sb.delete_mark(self.end_mark)
        
        self.sourceview = new_sv
        self.sourcebuffer = sb = new_sv.get_buffer()
        self.start_mark = sb.create_mark(None, sb.get_start_iter(),
                                         left_gravity=True)
        self.end_mark = sb.create_mark(None, sb.get_start_iter(),
                                       left_gravity=False)
        
    def connect(self, widget, *args):
        handler = widget.connect(*args)
        self.signals.append((widget, handler))

    def disconnect_all(self):
        for widget, handler in self.signals:
            widget.disconnect(handler)
        self.signals[:] = []

    def show(self, is_auto):
        sb = self.sourcebuffer
        text = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        index = sb.get_iter_at_mark(sb.get_insert()).get_offset()
        hp = HyperParser(text, index, self.INDENT_WIDTH)

        # This is used to write "return and_maybe_beep()".
        def and_maybe_beep():
            if not is_auto:
                beep()
            return None

        opener, closer = hp.get_surrounding_brackets('(')
        if not opener:
            return and_maybe_beep()
        if not closer:
            closer = len(text)
        hp.set_index(opener)
        expr = hp.get_expression()
        if not expr or (is_auto and expr.find('(') != -1):
            return and_maybe_beep()
        arg_text = self.get_func_doc(expr)

        if not arg_text:
            return and_maybe_beep()

        sb.move_mark(self.start_mark, sb.get_iter_at_offset(opener+1))
        sb.move_mark(self.end_mark, sb.get_iter_at_offset(closer))

        self.hide()

        x, y = self.get_position()
        self.ctwindow.show(arg_text, x, y)

        self.connect(sb, 'mark-set', self.on_mark_set)
        self.connect(sb, 'insert-text', self.on_insert_text)
        self.connect(sb, 'delete-range', self.on_delete_range)
        self.connect(self.sourceview, 'focus-out-event', self.on_focus_out)
        self.connect(self.window_main, 'configure-event', self.on_configure)

        self.is_shown = True

    def get_position(self):
        sv = self.sourceview
        sb = self.sourcebuffer

        insert_iter = sb.get_iter_at_mark(sb.get_insert())
        start_iter = sb.get_iter_at_mark(self.start_mark)
        start_iter.backward_chars(1)

        if insert_iter.get_line() == start_iter.get_line():
            it = start_iter
        else:
            it = insert_iter.copy()
            it.set_line_index(0)
        rect = sv.get_iter_location(it)
        x, y = rect.x, rect.y + rect.height
        x, y = sv.buffer_to_window_coords(gtk.TEXT_WINDOW_WIDGET, x, y)
        y = max(y, 0)
        sv_x, sv_y = sv.get_window(gtk.TEXT_WINDOW_WIDGET).get_origin()
        x += sv_x; y += sv_y

        return x, y
    
    def place_window(self):
        if not self.is_shown:
            # Was called as a callback, and window was already closed.
            return False
            
        x, y = self.get_position()
        self.ctwindow.move_perhaps(x, y)

        # Called by idle_add, don't call again.
        return False

    def on_mark_set(self, sb, it, mark):
        if mark is sb.get_insert():
            if (it.compare(sb.get_iter_at_mark(self.start_mark)) < 0
                or it.compare(sb.get_iter_at_mark(self.end_mark)) > 0):
                self.hide()
            else:
                idle_add(self.place_window)

    def on_insert_text(self, sb, it, text, _length):
        if ('(' in text
            or ')' in text
            or it.compare(sb.get_iter_at_mark(self.start_mark)) < 0
            or it.compare(sb.get_iter_at_mark(self.end_mark)) > 0):
            self.hide()
        else:
            idle_add(self.place_window)

    def on_delete_range(self, sb, start, end):
        text = get_text(sb, start, end)
        if ('(' in text
            or ')' in text
            or start.compare(sb.get_iter_at_mark(self.start_mark)) < 0
            or end.compare(sb.get_iter_at_mark(self.end_mark)) > 0):
            self.hide()
        else:
            idle_add(self.place_window)

    def on_focus_out(self, _widget, _event):
        self.hide()
    
    def on_configure(self, _widget, _event):
        self.hide()

    def hide(self):
        if not self.is_shown:
            return
        
        self.disconnect_all()
        self.ctwindow.hide()
        self.is_shown = False




# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['open_dialog', 'save_dialog']

"""
Easy to use wrappers around GTK file dialogs.
"""

import os
from os.path import abspath, dirname, basename, exists

import gtk

# Support translation in the future
_ = lambda s: s

def open_dialog(func, title, parent, filter_name, filter_pattern):
    """
    Display the Open dialog.
    func - a function which gets a file name and does something. If it throws
        an IOError, it will be catched and the user will get another chance.
    title - window title
    parent - parent window, or None
    filter_name - "HTML Files"
    filter_pattern - "*.html"
    """
    d = gtk.FileChooserDialog(
        title, parent,
        gtk.FILE_CHOOSER_ACTION_OPEN,
        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
         gtk.STOCK_OK, gtk.RESPONSE_OK))
    fil = gtk.FileFilter()
    fil.set_name(filter_name)
    fil.add_pattern(filter_pattern)
    d.add_filter(fil)
    while True:
        r = d.run()
        if r != gtk.RESPONSE_OK:
            break
        filename = abspath(d.get_filename().decode('utf8'))
        try:
            func(filename)
        except IOError, e:
            m = gtk.MessageDialog(d, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING,
                                    gtk.BUTTONS_OK)
            m.props.text = _('Error when loading file: %s') % e
            m.run()
            m.destroy()
        else:
            break
    d.destroy()

def save_dialog(func, title, parent, filter_name, filter_pattern, auto_ext=None,
                prev_dir=None, prev_name=None):
    """
    Display the Save As dialog.
    func - a function which gets a file name and does something. If it throws
        an IOError, it will be catched and the user will get another chance.
    title - window title
    parent - parent window, or None
    filter_name - "HTML Files"
    filter_pattern - "*.html"
    auto_ext - "html", if not None will be added if no extension given.
    prev_dir, prev_name - will set the default if given.
    
    Return True if file was saved.
    """
    d = gtk.FileChooserDialog(
        title, parent,
        gtk.FILE_CHOOSER_ACTION_SAVE,
        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
         gtk.STOCK_OK, gtk.RESPONSE_OK))
    fil = gtk.FileFilter()
    fil.set_name(filter_name)
    fil.add_pattern(filter_pattern)
    d.add_filter(fil)
    if prev_dir:
        d.set_current_folder(prev_dir)
    if prev_name:
        d.set_current_name(prev_name)
    saved = False
    while True:
        r = d.run()
        if r != gtk.RESPONSE_OK:
            break
        filename = abspath(d.get_filename()).decode('utf8')
        if auto_ext and not os.path.splitext(filename)[1]:
            filename += os.path.extsep + auto_ext
        if exists(filename):
            m = gtk.MessageDialog(d, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION)
            m.props.text = _('A file named "%s" already exists.  Do '
                                'you want to replace it?'
                                ) % basename(filename)
            m.props.secondary_text = _(
                'The file already exists in "%s".  Replacing it will '
                'overwrite its contents.'
                ) % basename(dirname(filename))
            m.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
            m.add_button(_('_Replace'), gtk.RESPONSE_OK)
            m.set_default_response(gtk.RESPONSE_CANCEL)
            mr = m.run()
            m.destroy()
            if mr == gtk.RESPONSE_CANCEL:
                continue
                
        try:
            func(filename)
        except IOError, e:
            m = gtk.MessageDialog(d, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING,
                                  gtk.BUTTONS_OK)
            m.props.text = _('Error when saving file: %s') % e
            m.run()
            m.destroy()
        else:
            saved = True
            break
    d.destroy()
    return saved


# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['CallTipWindow']

import gtk
from gtk import gdk
import pango
from gobject import TYPE_NONE

from .keyhandler import make_keyhandler_decorator, handle_keypress

N_ROWS = 4
N_COLS = 80

# A decorator for managing sourceview key handlers
keyhandlers = {}
keyhandler = make_keyhandler_decorator(keyhandlers)

class CallTipWindow(object):
    """
    This class manages the calltip window, which displays function documentation.
    The window is shown and hidden upon request.
    """
    # The window looks like this: Most of it is occupied by the text box.
    # Below we have a horizontal scroll bar, which is displayed only when
    # needed, and to the right there's a vertical scroll bar, which is always
    # displayed. Below it is a "resize grip", which lets you resize the window.
    # The window can be moved by dragging the main text area.
    
    # This looks pretty much like a ScrolledWindow, but a SW doesn't have the
    # resize grip, So we layout the widgets by ourselves, and handle scrolling.
    
    # We implement our own resize grip - for some reason, the resize grip of
    # a status bar doesn't work on popup windows.
    
    def __init__(self, sourceview, sv_changed):
        self.sourceview = sourceview
        sv_changed.append(self.on_sv_changed)
        
        # Widgets
        self.textview = tv = gtk.TextView()
        self.hscrollbar = hs = gtk.HScrollbar()
        self.vscrollbar = vs = gtk.VScrollbar()
        self.resizegrip = rg = gtk.EventBox()
        self.vbox1 = vb1 = gtk.VBox()
        self.vbox2 = vb2 = gtk.VBox()
        self.hbox = hb = gtk.HBox()
        self.window = win = gtk.Window(gtk.WINDOW_POPUP)
        
        self.char_width, self.char_height = self.get_char_size(tv)
        
        # Dragging vars
        self.is_dragging = None
        self.drag_x = None
        self.drag_y = None
        self.drag_left = None
        self.drag_top = None
        self.was_dragged = None
        
        # Resizing vars
        self.is_resizing = None
        self.resize_x = None
        self.resize_y = None
        self.resize_width = None
        self.resize_height = None
        
        self.was_displayed = False
        
        # Initialization
        
        style = gtk.rc_get_style_by_paths(
            tv.get_settings(), 'gtk-tooltip', 'gtk-tooltip', TYPE_NONE)
        tv.modify_text(gtk.STATE_NORMAL, style.fg[gtk.STATE_NORMAL])
        tv.modify_base(gtk.STATE_NORMAL, style.bg[gtk.STATE_NORMAL])
        tv.set_size_request(0,0)
        tv.props.editable = False
        
        tv.connect('event', self.on_textview_event)
        
        tv.set_scroll_adjustments(hs.props.adjustment, vs.props.adjustment)
        
        tv.connect('scroll-event', self.on_textview_scroll)
        
        hs.props.adjustment.connect('changed', self.on_hadj_changed)

        rg.add_events(gdk.BUTTON_PRESS_MASK
                      | gdk.BUTTON_MOTION_MASK
                      | gdk.BUTTON_RELEASE_MASK
                      | gdk.EXPOSURE_MASK)
        
        rg.connect('event', self.on_resizegrip_event)
        rg.set_size_request(vs.size_request()[0], vs.size_request()[0])
        
        vb1.pack_start(tv, True, True)
        vb1.pack_start(hs, False, False)
        vb2.pack_start(vs, True, True)
        vb2.pack_end(rg, False, False)
        hb.pack_start(vb1, True, True)
        hb.pack_start(vb2, False, False)
        win.add(hb)
        
        # Make all widgets except the window visible, so that a simple "show"
        # will suffice to show the window
        hb.show_all()
        
        # We define this handler here so that it will be defined before
        # the default key-press handler, and so will have higher priority.
        self.keypress_handler = self.sourceview.connect(
            'key-press-event', self.on_keypress)
        self.sourceview.handler_block(self.keypress_handler)
        self.keypress_handler_blocked = True

    def on_sv_changed(self, new_sv):
        self.hide()
        self.sourceview.disconnect(self.keypress_handler)
        self.sourceview = new_sv
        self.keypress_handler = self.sourceview.connect(
            'key-press-event', self.on_keypress)
        self.sourceview.handler_block(self.keypress_handler)

    @staticmethod
    def get_char_size(textview):
        """
        Get width, height of a character in pixels.
        """
        tv = textview
        context = tv.get_pango_context()
        metrics = context.get_metrics(tv.style.font_desc,
                                      context.get_language())
        width = pango.PIXELS(metrics.get_approximate_digit_width())
        height = pango.PIXELS(metrics.get_ascent() + metrics.get_descent())
        return width, height
    
    def on_textview_scroll(self, _widget, event):
        adj = self.vscrollbar.props.adjustment
        # Scrolling: 3 lines
        step = self.char_height * 3
        
        if event.direction == gtk.gdk.SCROLL_UP:
            adj.props.value -= step
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            adj.props.value = min(adj.props.value+step, 
                                  adj.props.upper-adj.props.page_size)

    def on_hadj_changed(self, adj):
        self.hscrollbar.props.visible = (adj.props.page_size < adj.props.upper)

    def on_textview_event(self, _widget, event):
        if event.type == gdk.BUTTON_PRESS:
            self.is_dragging = True
            self.was_dragged = True
            self.drag_x = event.x_root
            self.drag_y = event.y_root
            self.drag_left, self.drag_top = self.window.get_position()
            return True
        elif event.type == gdk.MOTION_NOTIFY and self.is_dragging:
            left = self.drag_left + event.x_root - self.drag_x
            top = self.drag_top + event.y_root - self.drag_y
            self.window.move(int(left), int(top))
            return True
        elif event.type == gdk.BUTTON_RELEASE:
            self.is_dragging = False

    def on_resizegrip_event(self, _widget, event):
        if event.type == gdk.BUTTON_PRESS:
            self.resize_x = event.x_root
            self.resize_y = event.y_root
            self.resize_width, self.resize_height = self.window.get_size()
            return True
        elif event.type == gdk.MOTION_NOTIFY:
            width = max(0, self.resize_width + event.x_root - self.resize_x)
            height = max(0, self.resize_height + event.y_root - self.resize_y)
            self.window.resize(int(width), int(height))
            return True
        elif event.type == gdk.EXPOSE:
            rg = self.resizegrip
            win = rg.window
            _x, _y, width, height, _depth = win.get_geometry()
            rg.get_style().paint_resize_grip(
                win, gtk.STATE_NORMAL, None, rg, None, 
                gdk.WINDOW_EDGE_SOUTH_EAST, 0, 0, width, height)
            return True
    
    @keyhandler('Up', 0)
    def on_up(self):
        adj = self.vscrollbar.props.adjustment
        adj.props.value -= self.char_height
        return True

    @keyhandler('Down', 0)
    def on_down(self):
        adj = self.vscrollbar.props.adjustment
        adj.props.value = min(adj.props.value + self.char_height, 
                              adj.props.upper - adj.props.page_size)
        return True

    @keyhandler('Page_Up', 0)
    def on_page_up(self):
        self.textview.emit('move-viewport', gtk.SCROLL_PAGES, -1)
        return True
        
    @keyhandler('Page_Down', 0)
    def on_page_down(self):
        self.textview.emit('move-viewport', gtk.SCROLL_PAGES, 1)
        return True

    @keyhandler('Escape', 0)
    def on_esc(self):
        self.hide()
        # Don't return True - other things may be escaped too.

    def on_keypress(self, _widget, event):
        return handle_keypress(self, event, keyhandlers)

    def show(self, text, x, y):
        """
        Show the window with the given text, its top-left corner at x-y.
        Decide on initial size.
        """
        # The initial size is the minimum of:
        # * N_COLS*N_ROWS
        # * Whatever fits into the screen
        # * The actual content
        
        tv = self.textview
        vs = self.vscrollbar
        win = self.window
        
        text = text.replace('\0', '') # Fixes bug #611513
        
        win.hide()
        tv.get_buffer().set_text(text)
        
        f_width = self.char_width * N_COLS
        f_height = self.char_height * N_ROWS
        
        s_width = gdk.screen_width() - x
        s_height = gdk.screen_height() - y
        
        # Get the size of the contents
        layout = tv.create_pango_layout(text)
        p_width, p_height = layout.get_size()
        c_width = pango.PIXELS(p_width)
        c_height = pango.PIXELS(p_height)
        del layout
        
        add_width = vs.size_request()[0] + 5
        width = int(min(f_width, s_width, c_width) + add_width)
        height = int(min(f_height, s_height, c_height))
        
        # Don't show the vertical scrollbar if the height is short enough.
        vs.props.visible = (height > vs.size_request()[1])
        
        win.resize(width, height)
        
        win.move(x, y)
        
        self.hscrollbar.props.adjustment.props.value = 0
        self.vscrollbar.props.adjustment.props.value = 0
        
        self.sourceview.handler_unblock(self.keypress_handler)
        self.keypress_handler_blocked = False

        win.show()
        
        # This has to be done after the textview was displayed
        if not self.was_displayed:
            self.was_displayed = True
            hand = gdk.Cursor(gdk.HAND1)
            tv.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(hand)
            br_corner = gdk.Cursor(gdk.BOTTOM_RIGHT_CORNER)
            self.resizegrip.window.set_cursor(br_corner)
    
    def hide(self):
        self.window.hide()

        if not self.keypress_handler_blocked:
            self.sourceview.handler_block(self.keypress_handler)
            self.keypress_handler_blocked = True

        self.is_dragging = False
        self.is_resizing = False
        self.was_dragged = False
    
    def move_perhaps(self, x, y):
        """
        Move the window to x-y, unless it was already manually dragged.
        """
        if not self.was_dragged:
            self.window.move(x, y)


# vim: sw=4 ts=4:
#
# (c) 2003 Gustavo J A M Carneiro gjc at inescporto.pt
#     2004-2005 Filip Van Raemdonck
#
# http://www.daa.com.au/pipermail/pygtk/2003-August/005775.html
# Message-ID: <1062087716.1196.5.camel@emperor.homelinux.net>
#     "The license is whatever you want."

import inspect, linecache, sys, traceback
from repr import repr as safe_repr
from cStringIO import StringIO
from gettext import gettext as _
#from smtplib import SMTP

import gtk
import pango

from .bug_report import bug_report

#def analyse (exctyp, value, tb):
#    trace = StringIO()
#    traceback.print_exception (exctyp, value, tb, None, trace)
#    return trace.getvalue()

def lookup (name, frame, lcls):
    '''Find the value for a given name in the given frame'''
    if name in lcls:
        return 'local', lcls[name]
    elif name in frame.f_globals:
        return 'global', frame.f_globals[name]
    elif '__builtins__' in frame.f_globals:
        builtins = frame.f_globals['__builtins__']
        if type (builtins) is dict:
            if name in builtins:
                return 'builtin', builtins[name]
        else:
            if hasattr (builtins, name):
                return 'builtin', getattr (builtins, name)
    return None, []

def analyse (exctyp, value, tb):
    import tokenize, keyword

    trace = StringIO()
    nlines = 1
    frecs = inspect.getinnerframes (tb, nlines)
    trace.write ('Variables:\n')
    for frame, fname, lineno, funcname, _context, _cindex in frecs:
        trace.write ('  File "%s", line %d, ' % (fname, lineno))
        args, varargs, varkw, lcls = inspect.getargvalues (frame)

        def readline (lno=[lineno], *args):
            if args: print args
            try: return linecache.getline (fname, lno[0])
            finally: lno[0] += 1
        all, prev, name, scope = {}, None, '', None
        for ttype, tstr, _stup, _etup, _line in tokenize.generate_tokens (readline):
            if ttype == tokenize.NAME and tstr not in keyword.kwlist:
                if name:
                    if name[-1] == '.':
                        try:
                            val = getattr (prev, tstr)
                        except AttributeError:
                            # XXX skip the rest of this identifier only
                            break
                        name += tstr
                else:
                    assert not name and not scope
                    scope, val = lookup (tstr, frame, lcls)
                    name = tstr
                if val is not None:
                    prev = val
                #print '  found', scope, 'name', name, 'val', val, 'in', prev, 'for token', tstr
            elif tstr == '.':
                if prev:
                    name += '.'
            else:
                if name:
                    all[name] = prev
                prev, name, scope = None, '', None
                if ttype == tokenize.NEWLINE:
                    break

        trace.write (funcname +
          inspect.formatargvalues (args, varargs, varkw, lcls, formatvalue=lambda v: '=' + safe_repr (v)) + '\n')
        if len (all):
            trace.write ('    %s\n' % str (all))

    trace.write('\n')
    traceback.print_exception (exctyp, value, tb, None, trace)
    
    return trace.getvalue()

def _info (exctyp, value, tb):
    # First output the exception to stderr
    orig_excepthook(exctyp, value, tb)
    
    try:
        import pdb
    except ImportError:
        # py2exe
        pdb = None
        
    if exctyp is KeyboardInterrupt:
        sys.exit(1)
    trace = None
    dialog = gtk.MessageDialog (parent=None, flags=0, type=gtk.MESSAGE_WARNING, buttons=gtk.BUTTONS_NONE)
    dialog.set_title (_("Bug Detected"))
    if gtk.check_version (2, 4, 0) is not None:
        dialog.set_has_separator (False)

    primary = _("<big><b>A programming error has been detected.</b></big>")
    secondary = _("Please report it by clicking the 'Report' button. Thanks!")

    dialog.set_markup (primary)
    dialog.format_secondary_text (secondary)

    dialog.add_button (_("Report..."), 3)
        
    dialog.add_button (_("Details..."), 2)
    if pdb:
        dialog.add_button (_("Debug..."), 4)
    dialog.add_button (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)

    while True:
        resp = dialog.run()
        if resp == 4:
            pdb.post_mortem(tb)
            
        if resp == 3:
            if trace == None:
                trace = analyse (exctyp, value, tb)
            
            bug_report(dialog, _gladefile, trace)

        elif resp == 2:
            if trace == None:
                trace = analyse (exctyp, value, tb)

            # Show details...
            details = gtk.Dialog (_("Bug Details"), dialog,
              gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
              (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE, ))
            details.set_property ("has-separator", False)

            textview = gtk.TextView(); textview.show()
            textview.set_editable (False)
            textview.modify_font (pango.FontDescription ("Monospace"))

            sw = gtk.ScrolledWindow(); sw.show()
            sw.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            sw.add (textview)
            details.vbox.add (sw)
            textbuffer = textview.get_buffer()
            textbuffer.set_text (trace)

            monitor = gtk.gdk.screen_get_default ().get_monitor_at_window (dialog.window)
            area = gtk.gdk.screen_get_default ().get_monitor_geometry (monitor)
            w = area.width // 1.6
            h = area.height // 1.6
            details.set_default_size (int (w), int (h))

            details.run()
            details.destroy()

        else:
            break

    dialog.destroy()

def install(gladefile):
    global orig_excepthook, _gladefile
    _gladefile = gladefile
    orig_excepthook = sys.excepthook
    sys.excepthook = _info

if __name__ == '__main__':
    class X (object):
        pass
    x = X()
    x.y = 'Test'
    x.z = x
    w = ' e'
    1, x.z.y, w
    raise Exception (x.z.y + w)




# This module is based on a code by Josiah Carlson,
# http://code.activestate.com/recipes/440554/
# Licensed under the PSF license.

import os
import subprocess
import errno
import time
import sys

PIPE = subprocess.PIPE

if subprocess.mswindows:
    from msvcrt import get_osfhandle #@UnresolvedImport
    from ctypes import byref, c_ulong, windll
    PeekNamedPipe = windll.kernel32.PeekNamedPipe #@UndefinedVariable
else:
    import select
    import fcntl

class Popen(subprocess.Popen):
    def recv(self, maxsize=None):
        return self._recv('stdout', maxsize)

    def recv_err(self, maxsize=None):
        return self._recv('stderr', maxsize)

    def send_recv(self, input='', maxsize=None):
        return self.send(input), self.recv(maxsize), self.recv_err(maxsize)

    def get_conn_maxsize(self, which, maxsize):
        if maxsize is None:
            maxsize = 1024
        elif maxsize < 1:
            maxsize = 1
        return getattr(self, which), maxsize

    def _close(self, which):
        getattr(self, which).close()
        setattr(self, which, None)

    if subprocess.mswindows:
        def send(self, input):
            if not self.stdin:
                return None

            try:
                written = os.write(self.stdin, input)
            except ValueError:
                return self._close('stdin')
            except (subprocess.pywintypes.error, Exception), why:
                if why[0] in (109, errno.ESHUTDOWN):
                    return self._close('stdin')
                raise

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            read = ""
            if conn is None:
                return None

            try:
                fd = conn.fileno()
                handle = get_osfhandle(fd)
                avail = c_ulong(0)
                PeekNamedPipe(handle, None, 0, None, byref(avail), None)
                nAvail = avail.value
                if maxsize < nAvail:
                    nAvail = maxsize
                if nAvail > 0:
                    read = os.read(fd, nAvail)
            except ValueError:
                return self._close(which)
            except (subprocess.pywintypes.error, Exception), why:
                if why[0] in (109, errno.ESHUTDOWN):
                    return self._close(which)
                raise

            if self.universal_newlines:
                read = self._translate_newlines(read)
            return read

    else:
        def send(self, input):
            if not self.stdin:
                return None

            if not select.select([], [self.stdin], [], 0)[1]:
                return 0

            try:
                written = os.write(self.stdin.fileno(), input)
            except OSError, why:
                if why[0] == errno.EPIPE: #broken pipe
                    return self._close('stdin')
                raise

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            flags = fcntl.fcntl(conn, fcntl.F_GETFL)
            if not conn.closed:
                fcntl.fcntl(conn, fcntl.F_SETFL, flags| os.O_NONBLOCK)

            try:
                if not select.select([conn], [], [], 0)[0]:
                    return ''

                r = conn.read(maxsize)
                if not r:
                    return self._close(which)

                if self.universal_newlines:
                    r = self._translate_newlines(r)
                return r
            finally:
                if not conn.closed:
                    fcntl.fcntl(conn, fcntl.F_SETFL, flags)

message = "Other end disconnected!"

def recv_some(p, t=.1, e=1, tr=5, stderr=0):
    if tr < 1:
        tr = 1
    x = time.time()+t
    y = []
    r = ''
    pr = p.recv
    if stderr:
        pr = p.recv_err
    while time.time() < x or r:
        r = pr()
        if r is None:
            if e:
                raise Exception(message)
            else:
                break
        elif r:
            y.append(r)
        else:
            time.sleep(max((x-time.time())/tr, 0))
    return ''.join(y)

def send_all(p, data):
    while len(data):
        sent = p.send(data)
        if sent is None:
            raise Exception(message)
        data = buffer(data, sent)

if __name__ == '__main__':
    if sys.platform == 'win32':
        shell, commands, tail = ('cmd', ('dir /w', 'echo HELLO WORLD'), '\r\n')
    else:
        shell, commands, tail = ('sh', ('ls', 'echo HELLO WORLD'), '\n')

    a = Popen(shell, stdin=PIPE, stdout=PIPE)
    print recv_some(a),
    for cmd in commands:
        send_all(a, cmd + tail)
        print recv_some(a),
    send_all(a, 'exit' + tail)
    print recv_some(a, e=0)
    a.wait()




# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['HistPersist']

import os
from HTMLParser import HTMLParser
from htmlentitydefs import name2codepoint

from .file_dialogs import open_dialog, save_dialog
from .common import get_text

_ = lambda s: s

class HistPersist(object):
    """
    Provide actions for storing and loading history.
    """
    
    def __init__(self, window_main, textview, status_bar, recent_manager):
        self.window_main = window_main
        self.textview = textview
        self.textbuffer = textview.get_buffer()
        self.status_bar = status_bar
        self.recent_manager = recent_manager
        
        self.filename = None
        
        self.textbuffer.connect('modified-changed', self.on_modified_changed)
    
    def save_filename(self, filename):
        """
        Save history to a file.
        """
        f = open(filename, 'wb')
        save_history(self.textview, f)
        f.close()
        self.filename = filename
        self.status_bar.set_status(_('History saved.'))
        self.recent_add(filename)
        self.textbuffer.set_modified(False)

    def save(self):
        """
        Show the save dialog if there's no filename. Return True if was saved.
        """
        if self.filename is None:
            saved = self.save_as()
        else:
            self.save_filename(self.filename)
            saved = True
        return saved
    
    def save_as(self):
        """Show the save dialog. Return True if was saved."""
        if self.filename:
            prev_dir = os.path.dirname(self.filename)
            prev_name = os.path.basename(self.filename)
        else:
            prev_dir = None
            #prev_name = 'dreampie-history.html'
            prev_name = None
        saved = save_dialog(self.save_filename,
                            _('Choose where to save the history'),
                            self.window_main,
                            _('HTML Files'),
                            '*.html', 'html',
                            prev_dir, prev_name)
        return saved

    def load_filename(self, filename):
        s = open(filename, 'rb').read()
        parser = Parser(self.textbuffer)
        parser.feed(s)
        parser.close()
        self.status_bar.set_status(_('History loaded.'))
        self.filename = filename
        self.update_title()
        self.recent_add(filename)
    
    def load(self):
        open_dialog(self.load_filename,
                    _('Choose the saved history file'),
                    self.window_main,
                    _('HTML Files'),
                    '*.html')
    
    def recent_add(self, filename):
        # FIXME: This doesn't add an entry when saving HTML files. VERY strange.
        self.recent_manager.add_full('file://'+filename, {
            'mime_type': 'text/html', 'app_name': 'dreampie',
            'app_exec': 'dreampie'})
    
    def update_title(self):
        if self.filename:
            disp_fn = os.path.basename(self.filename)
            if self.textbuffer.get_modified():
                disp_fn += '*'
            self.window_main.set_title("%s - DreamPie" % disp_fn)
        else:
            self.window_main.set_title("DreamPie")
    
    def on_modified_changed(self, _widget):
        if self.filename:
            self.update_title()
    
    def was_saved(self):
        return self.filename is not None
    
    def forget_filename(self):
        self.filename = None
        self.update_title()


def _html_escape(s):
    """
    Replace special characters "&", "<" and ">" to HTML-safe sequences.
    """
    # This is taken from cgi.escape - I didn't want to import it, because of
    # py2exe
    s = s.replace("&", "&amp;") # Must be done first!
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    return s

def _format_color(color):
    return '#%02x%02x%02x' % (color.red >> 8, color.green >> 8, color.blue >> 8)

def save_history(textview, f):
    """
    Save the history - the content of the textview - to a HTML file f.
    """
    tv = textview
    tb = tv.get_buffer()
    style = tv.get_style()

    f.write("""\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN">
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<meta name="DreamPie Format" content="1">
<title>DreamPie History</title>
<style>
body {
  white-space: pre-wrap;
  font-family: %s;
  font-size: %s;
  color: %s;
  background-color: %s;
}
""" % (
    style.font_desc.get_family(),
    style.font_desc.get_size(),
    _format_color(style.text[0]),
    _format_color(style.base[0]),
    )
)
    
    tt = tb.get_tag_table()
    all_tags = []
    tt.foreach(lambda tag, _data: all_tags.append(tag))
    all_tags.sort(key=lambda tag: -tag.get_priority())
    
    for tag in all_tags:
        f.write("span.%s {\n" % tag.props.name)
        if tag.props.foreground_set:
            f.write("  color: %s;\n" % _format_color(tag.props.foreground_gdk))
        if tag.props.background_set:
            f.write("  background-color: %s;\n"
                    % _format_color(tag.props.background_gdk))
        if tag.props.invisible:
            f.write(" display: none;\n")
        f.write("}\n")
    
    f.write("""\
</style>
</head>
<body>""")
    
    cur_tags = []
    it = tb.get_start_iter()
    while True:
        new_tags = cur_tags[:]
        for tag in it.get_toggled_tags(False):
            new_tags.remove(tag)
        for tag in it.get_toggled_tags(True):
            new_tags.append(tag)
        new_tags.sort(key=lambda tag: -tag.get_priority())
        
        shared_prefix = 0
        while (len(cur_tags) > shared_prefix and len(new_tags) > shared_prefix
               and cur_tags[shared_prefix] is new_tags[shared_prefix]):
            shared_prefix += 1
        for _i in range(len(cur_tags) - shared_prefix):
            f.write('</span>')
        for tag in new_tags[shared_prefix:]:
            f.write('<span class="%s">' % tag.props.name)
        
        if it.compare(tb.get_end_iter()) == 0:
            # We reached the end. We break here, because we want to close
            # the tags.
            break
        
        new_it = it.copy()
        new_it.forward_to_tag_toggle(None)
        text = get_text(tb, it, new_it)
        text = _html_escape(text)
        f.write(text.encode('utf8'))
        
        it = new_it
        cur_tags = new_tags
    
    f.write("""\
</body>
</html>
""")

class LoadError(Exception):
    pass

class Parser(HTMLParser):
    def __init__(self, textbuffer):
        HTMLParser.__init__(self)
        
        self.textbuffer = tb = textbuffer

        self.reached_body = False
        self.version = None
        self.cur_tags = []
        self.leftmark = tb.create_mark(None, tb.get_start_iter(), True)
        self.rightmark = tb.create_mark(None, tb.get_start_iter(), False)
    
    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if not self.reached_body:
            if tag == 'meta':
                if 'name' in attrs and attrs['name'] == 'DreamPie Format':
                    if attrs['content'] != '1':
                        raise LoadError("Unrecognized DreamPie Format")
                    self.version = 1
            if tag == 'body':
                if self.version is None:
                    raise LoadError("File is not a DreamPie history file.")
                self.reached_body = True
        else:
            if tag == 'span':
                if 'class' not in attrs:
                    raise LoadError("<span> without a 'class' attribute")
                self.cur_tags.append(attrs['class'])
    
    def handle_endtag(self, tag):
        if tag == 'span':
            if not self.cur_tags:
                raise LoadError("Too many </span> tags")
            self.cur_tags.pop()
    
    def insert(self, data):
        tb = self.textbuffer
        leftmark = self.leftmark; rightmark = self.rightmark
        # For some reasoin, insert_with_tags_by_name marks everything with the
        # message tag. So we do it all by ourselves...
        tb.insert(tb.get_iter_at_mark(leftmark), data)
        leftit = tb.get_iter_at_mark(leftmark)
        rightit = tb.get_iter_at_mark(rightmark)
        tb.remove_all_tags(leftit, rightit)
        for tag in self.cur_tags:
            tb.apply_tag_by_name(tag, leftit, rightit)
        tb.move_mark(leftmark, rightit)

    def handle_data(self, data):
        if self.reached_body:
            self.insert(data.decode('utf8'))
    
    def handle_charref(self, name):
        raise LoadError("Got a charref %r and not expecting it." % name)
    
    def handle_entityref(self, name):
        if self.reached_body:
            self.insert(unichr(name2codepoint[name]))
    
    def close(self):
        HTMLParser.close(self)
        
        tb = self.textbuffer
        tb.delete_mark(self.leftmark)
        tb.delete_mark(self.rightmark)



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Output']

import re
from StringIO import StringIO

from .tags import OUTPUT
from .common import get_text

# This RE is used to remove chars that won't be displayed from the data string.
remove_cr_re = re.compile(r'\n[^\n]*\r')
# Match ANSI escapes. See http://en.wikipedia.org/wiki/ANSI_escape_code
ansi_escape_re = re.compile(r'\x1b\[[^@-~]*?[@-~]')

# Length after which to break a line with a '\r' - a character which we
# ignore when copying.
BREAK_LEN = 1600

class Output(object):
    """
    Manage writing output to the text view.
    See a long documentation string in tags.py for more information about the
    model.
    """
    def __init__(self, textview):
        self.textview = textview
        self.textbuffer = tb = textview.get_buffer()

        # A mark where new output should be written
        self.mark = tb.create_mark(None, tb.get_end_iter(), left_gravity=True)
        # If the real output doesn't end with a newline, we add "our own",
        # because we want the output section to always end with a newline.
        # This newline will be deleted if more output is written.
        # If we did, self.added_newline is True.
        self.added_newline = False
        # Was something written at all in this section?
        self.was_something_written = False
        # Does the output end with a cr? (If it does, the last line will be
        # deleted unless the next output starts with a lf)
        self.is_cr = False

    def start_new_section(self):
        tb = self.textbuffer
        it = tb.get_end_iter()
        tb.move_mark(self.mark, it)
        self.added_newline = False
        self.was_something_written = False
        self.is_cr = False

    def write(self, data, tag_names, onnewline=False, addbreaks=True):
        """
        Write data (unicode string) to the text buffer, marked with tag_names.
        (tag_names can be either a string or a list of strings)
        If onnewline is True, will add a newline if the output until now doesn't
        end with one.
        If addbreaks is True, '\r' chars will be added so that lines will be
        broken and output will not burden the textview.
        Return a TextIter pointing to the end of the written text.
        """
        tb = self.textbuffer
        
        if isinstance(tag_names, basestring):
            tag_names = [tag_names]
        
        if not data:
            return
        
        if self.added_newline:
            if onnewline:
                # If we added a newline, it means that the section didn't end
                # with a newline, so we need to add one.
                data = '\n' + data
            it = tb.get_iter_at_mark(self.mark)
            it2 = it.copy()
            it2.backward_char()
            assert get_text(tb, it2, it) == '\n'
            tb.delete(it2, it)
            self.added_newline = False

        # Keep lines if after the cr there was no data before the lf.
        # Since that's the normal Windows newline, it's very important.
        data = data.replace('\r\n', '\n')
        
        # Remove ANSI escapes
        data = ansi_escape_re.sub('', data)
        
        # Remove NULL chars
        data = data.replace('\0', '')
        
        has_trailing_cr = data.endswith('\r')
        if has_trailing_cr:
            data = data[:-1]
        
        if data.startswith('\n'):
            # Don't delete the last line if it ended with a cr but this data
            # starts with a lf.
            self.is_cr = False
            
        # Remove chars that will not be displayed from data. No crs will be left
        # after the first lf.
        data = remove_cr_re.sub('\n', data)

        cr_pos = data.rfind('\r')
        if (self.is_cr or cr_pos != -1) and self.was_something_written:
            # Delete last written line
            it = tb.get_iter_at_mark(self.mark)
            output_start = it.copy()
            output_tag = tb.get_tag_table().lookup(OUTPUT)
            output_start.backward_to_tag_toggle(output_tag)
            assert output_start.begins_tag(output_tag)
            r = it.backward_search('\n', 0, output_start)
            if r is not None:
                _before_newline, after_newline = r
            else:
                # Didn't find a newline - delete from beginning of output
                after_newline = output_start
            tb.delete(after_newline, it)

        # Remove data up to \r.
        if cr_pos != -1:
            data = data[cr_pos+1:]

        if addbreaks:
            # We DO use \r characters as linebreaks after BREAK_LEN chars, which
            # are not copied.
            f = StringIO()
    
            pos = 0
            copied_pos = 0
            col = tb.get_iter_at_mark(self.mark).get_line_offset()
            next_newline = data.find('\n', pos)
            if next_newline == -1:
                next_newline = len(data)
            while pos < len(data):
                if next_newline - pos + col > BREAK_LEN:
                    pos = pos + BREAK_LEN - col
                    f.write(data[copied_pos:pos])
                    f.write('\r')
                    copied_pos = pos
                    col = 0
                else:
                    pos = next_newline + 1
                    col = 0
                    next_newline = data.find('\n', pos)
                    if next_newline == -1:
                        next_newline = len(data)
            f.write(data[copied_pos:])
            data = f.getvalue()

        it = tb.get_iter_at_mark(self.mark)
        tb.insert_with_tags_by_name(it, data, OUTPUT, *tag_names)

        if not data.endswith('\n'):
            tb.insert_with_tags_by_name(it, '\n', OUTPUT)
            self.added_newline = True
        
        # Move mark to after the written text
        tb.move_mark(self.mark, it)

        self.is_cr = has_trailing_cr

        self.was_something_written = True
        
        return it





# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['beep', 'get_text']

import sys
if sys.platform == 'win32':
    from winsound import MessageBeep as beep #@UnresolvedImport @UnusedImport
else:
    from gtk.gdk import beep #@UnusedImport @Reimport

def get_text(textbuffer, *args):
    # Unfortunately, PyGTK returns utf-8 encoded byte strings instead of unicode
    # strings. There's no point in getting the utf-8 byte string, so whenever
    # TextBuffer.get_text is used, this function should be used instead.
    return textbuffer.get_text(*args).decode('utf8')

class TimeoutError(Exception):
    pass



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

# This file is a script (not a module) run by the DreamPie GUI.
# It expects one argument: the port to connect to.
# It creates a package called dreampielib from subp-py2.zip or subp-py3.zip
# (which are expected to be in the directory of __file__),
# and runs dreampielib.subprocess.main(port).

# This is a hack to solve bug #527630. Python2.5 ignores the PYTHONIOENCODING
# environment variable, but we want to set the output encoding to utf-8 so that
# unicode chars will be printed. So we disable automatic loading of site.py with
# the -S flag, and call sys.setdefaultencoding before site.py has a chance of
# doing anything else.
import sys
if sys.version_info[0] < 3:
    sys.setdefaultencoding('utf-8') #@UndefinedVariable
import site #@UnusedImport

from os.path import abspath, join, dirname

def main():
    port = int(sys.argv[1])

    py_ver = sys.version_info[0]
    lib_name = abspath(join(dirname(__file__), 'subp-py%d' % py_ver))
    
    sys.path.insert(0, lib_name)
    from dreampielib.subprocess import main as subprocess_main
    del sys.path[0]
    
    if sys.version_info[:2] == (3, 0):
        sys.stderr.write("Warning: DreamPie doesn't support Python 3.0. \n"
                         "Please upgrade to Python 3.1.\n")
    
    subprocess_main(port)

if __name__ == '__main__':
    main()



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['trunc_traceback']

import sys
py3k = (sys.version_info[0] == 3)
import traceback
import linecache
from StringIO import StringIO

def unicodify(s):
    """Fault-tolerant conversion to unicode"""
    return s if isinstance(s, unicode) else s.decode('utf8', 'replace')

#######################################################################
# This is copied from traceback.py from Python 3.1.1.
# It is copied because I don't want to rely on private functions.

_cause_message = (
    "\nThe above exception was the direct cause "
    "of the following exception:\n")

_context_message = (
    "\nDuring handling of the above exception, "
    "another exception occurred:\n")

def _iter_chain(exc, custom_tb=None, seen=None):
    if seen is None:
        seen = set()
    seen.add(exc)
    its = []
    cause = exc.__cause__
    context = exc.__context__
    if cause is not None and cause not in seen:
        its.append(_iter_chain(cause, None, seen))
        its.append([(_cause_message, None)])
    if context is not None and context is not cause and context not in seen:
        its.append(_iter_chain(context, None, seen))
        its.append([(_context_message, None)])
    its.append([(exc, custom_tb or exc.__traceback__)])
    # itertools.chain is in an extension module and may be unavailable
    for it in its:
        for x in it:
            yield x

# Copied up to here.
#######################################################################


def canonical_fn(fn):
    """
    Return something that will be equal for both source file and the cached
    compile file.
    """
    # If the file contains a '$', remove from it (Jython uses it). Otherwise,
    # remove from a '.'.
    if '$' in fn:
        return fn.rsplit('$', 1)[0]
    else:
        return fn.rsplit('.', 1)[0]

def trunc_traceback((_typ, value, tb), source_file):
    """
    Format a traceback where entries before a frame from source_file are
    omitted (unless the last frame is from source_file).
    Return the result as a unicode string.
    """
    # This is complicated because we want to support nested tracebacks
    # in Python 3.

    linecache.checkcache()
    efile = StringIO()
    
    if py3k:
        values = _iter_chain(value, tb)
    else:
        values = [(value, tb)]
    
    # The source_file and filename may differ in extension (pyc/py), so we
    # ignore the extension
    source_file = canonical_fn(source_file)
    
    for value, tb in values:
        if isinstance(value, basestring):
            efile.write(value+'\n')
            continue
    
        tbe = traceback.extract_tb(tb)
        # This is a work around a really weird IronPython bug.
        while len(tbe)>1 and 'split_to_singles' in tbe[-1][0]:
            tbe.pop()
            
        # tbe may be an empty list if "raise from ExceptionClass" was used.
        if tbe and canonical_fn(tbe[-1][0]) != source_file:
            # If the last entry is from this file, don't remove
            # anything. Otherwise, remove lines before the current
            # frame.
            for i in xrange(len(tbe)-2, -1, -1):
                if canonical_fn(tbe[i][0]) == source_file:
                    tbe = tbe[i+1:]
                    break
                
        if tbe:
            efile.write('Traceback (most recent call last):'+'\n')
        traceback.print_list(tbe, file=efile)
        lines = traceback.format_exception_only(type(value), value)
        for line in lines:
            efile.write(line)
            
    if not hasattr(efile, 'buflist'):
        # Py3k
        return efile.getvalue()
    else:
        # The following line replaces efile.getvalue(), because if it
        # includes both unicode strings and byte string with non-ascii
        # chars, it fails.
        return u''.join(unicodify(s) for s in efile.buflist)



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

"""
An ugly hack to hide the console window associated with the current process.
See http://support.microsoft.com/kb/124103
"""

__all__ = ['hide_console_window']

import time
import ctypes
kernel32 = ctypes.windll.kernel32
user32 = ctypes.windll.user32

def hide_console_window():
    BUFSIZE = 1024
    buf = ctypes.create_string_buffer('', BUFSIZE)

    # Get current title
    length = kernel32.GetConsoleTitleA(buf, BUFSIZE)
    title = buf.raw[:length]

    # Change title to a unique string
    temp_title = '%s/%s' % (kernel32.GetCurrentProcessId(),
                            kernel32.GetTickCount())
    kernel32.SetConsoleTitleA(temp_title)
    time.sleep(.04)

    # Get window handle
    handle = user32.FindWindowA(None, temp_title)

    # Get current title, to make sure that we got the right handle
    length = user32.GetWindowTextA(handle, buf, BUFSIZE)
    cur_title = buf.raw[:length]

    # Restore title
    kernel32.SetConsoleTitleA(title)
    
    if cur_title == temp_title:
        # We got the correct handle, so hide the window.
        SW_HIDE = 0
        user32.ShowWindow(handle, SW_HIDE)
        



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Autoparen']

import string
from keyword import iskeyword
import re

from .hyper_parser import HyperParser
from .common import get_text

# These are all the chars that may be before the parens
LAST_CHARS = set(string.ascii_letters + string.digits + "_)]")

# Compile REs for checking if we are between 'for' and 'in'.
_for_re = re.compile(r'\bfor\b')
_in_re = re.compile(r'\bin\b')

# If after adding parens one of these strings is typed, we remove the parens.
# These are symbols (or prefixes of symbols) which don't make sense at the
# beginning of an expression, but do between two expressions - for example,
# "a and b" is fine, but "a(and b)" doesn't make sense.
undo_strings = set(x+' ' for x in 'and or is not if else for as'.split())
undo_strings.update('! % & * + / < = > ^ , ) ] }'.split())
undo_strings.add('- ') # A binary '-', not an unary '-'.
# A set of prefixes of undo_strings
prefixes = set(s[:i] for s in undo_strings for i in range(1, len(s)))

class Autoparen(object):
    """
    Add parentheses if a space was pressed after a callable-only object.
    """

    def __init__(self, sourcebuffer, sv_changed, is_callable_only, get_expects_str,
                 show_call_tip, INDENT_WIDTH):
        self.sourcebuffer = sb = sourcebuffer
        sv_changed.append(self.on_sv_changed)
        self.is_callable_only = is_callable_only
        self.get_expects_str = get_expects_str
        self.show_call_tip = show_call_tip
        self.INDENT_WIDTH = INDENT_WIDTH
        
        # We place this mark at the end of the expression we added parens to,
        # so that if the user removes the paren and presses space, we won't
        # interfere another time.
        self.mark = sb.create_mark(None, sb.get_start_iter(), left_gravity=True)
        
        # If a string in undo_strings is typed, we undo. We track changes to
        # the sourcebuffer until we are not in 'prefixes' or we are in
        # undo_strings.
        # To accomplish that, we listen for insert-text and delete-range
        # signals while we are in 'prefixes'.
        self.cur_prefix = None
        self.insert_handler = None
        self.delete_handler = None

    def on_sv_changed(self, new_sv):
        self.sourcebuffer.delete_mark(self.mark)
        self.disconnect()
        self.sourcebuffer = sb = new_sv.get_buffer()
        self.mark = sb.create_mark(None, sb.get_start_iter(), left_gravity=True)
    
    def add_parens(self):
        """
        This is called if the user pressed space on the sourceview, and
        the subprocess is not executing commands (so is_callable_only can work.)
        Should return True if event-handling should stop, or False if it should
        continue as usual.
        """
        sb = self.sourcebuffer
        
        # Quickly discard some cases
        insert = sb.get_iter_at_mark(sb.get_insert())
        mark_it = sb.get_iter_at_mark(self.mark)
        if mark_it.equal(insert):
            return False
        it = insert.copy()
        it.backward_char()
        if it.get_char() not in LAST_CHARS:
            return False
        it.forward_char()
        it.backward_word_start()
        if iskeyword(get_text(sb, it, insert)):
            return False
        
        text = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        index = sb.get_iter_at_mark(sb.get_insert()).get_offset()

        line = text[text.rfind('\n', 0, index)+1:index].lstrip()
        # don't add parens in import and except statements
        if line.startswith(('import ', 'from ', 'except ')):
            return False
        # don't add parens between 'for' and 'in'
        m = list(_for_re.finditer(line))
        if m:
            if not _in_re.search(line, m[-1].end()):
                return False

        hp = HyperParser(text, index, self.INDENT_WIDTH)

        if not hp.is_in_code():
            return False

        expr = hp.get_expression()
        if not expr:
            return False
        if '(' in expr:
            # Don't evaluate expressions which may contain a function call.
            return False
        
        r = self.is_callable_only(expr)
        if r is None:
            return False
        is_callable_only, expects_str = r
        if not is_callable_only:
            return False
        
        sb.move_mark(self.mark, insert)
        
        last_name = expr.rsplit('.', 1)[-1]
        sb.begin_user_action()
        if expects_str or last_name in self.get_expects_str():
            sb.insert(insert, '("")')
            insert.backward_chars(2)
        else:
            sb.insert(insert, '()')
            insert.backward_char()
        sb.place_cursor(insert)
        sb.end_user_action()
        
        if not expects_str:
            self.cur_prefix = ''
            self.disconnect()
            self.insert_handler = sb.connect('insert-text', self.on_insert_text)
            self.delete_handler = sb.connect('delete-range', self.on_delete_range)

        self.show_call_tip()
        
        return True
    
    def disconnect(self):
        if self.insert_handler:
            self.sourcebuffer.disconnect(self.insert_handler)
            self.insert_handler = None
        if self.delete_handler:
            self.sourcebuffer.disconnect(self.delete_handler)
            self.delete_handler = None
    
    def on_insert_text(self, _textbuffer, iter, text, _length):
        sb = self.sourcebuffer
        
        if len(text) != 1:
            self.disconnect()
            return
        it = sb.get_iter_at_mark(self.mark)
        it.forward_chars(len(self.cur_prefix)+1)
        if not it.equal(iter):
            self.disconnect()
            return
        
        new_prefix = self.cur_prefix + text
        if new_prefix in prefixes:
            # We continue to wait
            self.cur_prefix = new_prefix
        elif new_prefix in undo_strings:
            # Undo adding the parens.
            # Currently we have: "obj(prefi|)"
            # ("|" is iter. The last char wasn't written yet.)
            # We want: "obj prefix".
            # (the last char will be added by the default event handler.)
            # So we delete '(' and ')' and insert ' '.
            # We must keep 'iter' validated for the default handler, so it is
            # used in all insert and delete operations.
            self.disconnect()
            it = iter.copy()
            it.forward_char()
            sb.delete(iter, it)
            iter.backward_chars(len(self.cur_prefix))
            it = iter.copy()
            it.backward_char()
            sb.delete(it, iter)
            sb.insert(iter, ' ')
            iter.forward_chars(len(self.cur_prefix))
            return
        else:
            self.disconnect()
    
    def on_delete_range(self, _textbuffer, start, end):
        sb = self.sourcebuffer
        it = sb.get_iter_at_mark(self.mark)
        it.forward_chars(len(self.cur_prefix))
        it2 = it.copy()
        it2.forward_char()
        if self.cur_prefix and it.equal(start) and it2.equal(end):
            # BS was pressed, remove a char from cur_prefix and keep watching.
            self.cur_prefix = self.cur_prefix[:-1]
        else:
            self.disconnect()


# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['VAdjToBottom']

try:
    from glib import idle_add
except ImportError:
    # In PyGObject 2.14, it's in gobject.
    from gobject import idle_add

class VAdjToBottom(object):
    """
    Scroll automatically to the bottom of the VAdj if the height changes
    and it was at bottom
    """
    # The way this works is a little bit tricky, so here is the reasoning:
    # self.user_wants_bottom records whether the user wants to see the bottom,
    # so we should automatically scroll if more text was added. It is changed
    # by self.on_value_changed; It assumes that if the scrollbar is at the
    # bottom then that's what the user wants, but if it isn't at the bottom,
    # it means that the user doesn't want to see the bottom only if the
    # scrollbar was scrolled upwards - otherwise it's just us trying to catch
    # up.
    # self.on_changed monitors changes in the textview. If the scrollbar isn't
    # at the bottom (as the result of changes) but self.user_wants_bottom is
    # True, it schedules a call to scroll_to_bottom when idle. We don't call
    # scroll_to_bottom immediately because many times the are a few changes
    # before display, and scrolling before they are finished will cause
    # redisplay after every stage, which will be slow.
    def __init__(self, vadj):
        self.vadj = vadj
        self.last_value = self.vadj.value
        self.user_wants_bottom = True
        self.is_scroll_scheduled = False
        
        vadj.connect('changed', self.on_changed)
        vadj.connect('value-changed', self.on_value_changed)

    def is_at_bottom(self):
        return self.vadj.value + self.vadj.page_size - self.vadj.upper == 0

    def scroll_to_bottom(self):
        # Callback function
        try:
            self.is_scroll_scheduled = False
            self.vadj.set_value(self.vadj.upper - self.vadj.page_size)
        finally:
            # Avoid future calls
            return False

    def on_changed(self, _widget):
        if (not self.is_scroll_scheduled
            and self.user_wants_bottom
            and not self.is_at_bottom()):
            idle_add(self.scroll_to_bottom)
            self.is_scroll_scheduled = True

    def on_value_changed(self, _widget):
        is_at_bottom = self.is_at_bottom()
        if is_at_bottom:
            self.user_wants_bottom = True
        else:
            if self.vadj.value < self.last_value:
                self.user_wants_bottom = False
        self.last_value = self.vadj.value



# This file was automatically generated from a file in the source DreamPie dir.
# DO NOT EDIT IT, as your changes will be gone when the file is created again.

# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.


import sys
import collections
py3k = (sys.version_info[0] == 3)
import os
import time
import socket
from select import select
from io import StringIO
import linecache
import traceback
import types
import keyword
import builtins
import inspect
import pydoc
import pprint
import codeop
import signal
from contextlib import contextmanager
from itertools import chain
try:
    # Executing multiple statements in 'single' mode (print results) is done
    # with the ast module. Python 2.5 doesn't have it, so we use the compiler
    # module, which also seems to work. Jython 2.5 does have the ast module,
    # which is fortunate, because the 'compiler module trick' doesn't work there.
    import ast
except ImportError:
    ast = None
else:
    # IronPython 2.7.1 has the ast module, but can't compile from an AST.
    # If that's the case, the ast module doesn't interest us.
    try:
        compile(compile('a', 'fn', 'exec', ast.PyCF_ONLY_AST), 'fn', 'exec')
    except TypeError:
        ast = None
if ast is None:
    from .split_to_singles import split_to_singles
import __future__

if sys.platform == 'win32':
    from msvcrt import get_osfhandle #@UnresolvedImport
    from ctypes import byref, c_ulong, windll
    PeekNamedPipe = windll.kernel32.PeekNamedPipe #@UndefinedVariable

from .trunc_traceback import trunc_traceback
from .find_modules import find_modules
# We don't use relative import because of a Jython 2.5.1 bug.
from dreampielib.common.objectstream import send_object, recv_object

#import rpdb2; rpdb2.start_embedded_debugger('a')

import logging
from logging import debug
#logging.basicConfig(filename='/tmp/dreampie_subp_log', level=logging.DEBUG)

# time interval to process GUI events, in seconds
GUI_SLEEP = 0.1

# Maximum result string length to transmit
MAX_RES_STR_LEN = 1000000

rpc_funcs = set()
# A decorator which adds the function name to rpc_funcs
def rpc_func(func):
    rpc_funcs.add(func.__name__)
    return func

# Taken from codeop.py
PyCF_DONT_IMPLY_DEDENT = 0x200
_features = [getattr(__future__, fname)
             for fname in __future__.all_feature_names]

case_insen_filenames = (os.path.normcase('A') == 'a')

def unicodify(s):
    """Fault-tolerant conversion to unicode"""
    return s if isinstance(s, str) else s.decode('utf8', 'replace')

class PlainTextDoc(pydoc.TextDoc):
    """pydoc.TextDoc returns strange bold text, so we disable it."""
    def bold(self, text):
        return text
textdoc = PlainTextDoc()

# A mapping from id of types to boolean: is the type callable only.
# We use ids instead of weakrefs because old classes aren't weakrefable,
# and because we don't want to rely on weakref. This will fail if a type
# was deleted and another was created at the same memory address, but this
# seems unlikely.
is_callable_cache = {}
# magic methods for objects with defined operators
operator_methods = ['__%s__' % s for s in
                    'add sub mul div floordiv truediv mod divmod pow lshift '
                    'rshift and xor or'.split()]

quit_msg = """\
Press Ctrl-Q or close the window if you want to quit DreamPie.
Press Ctrl-F6 if you want to restart the subprocess."""
class Quit(object):
    def __repr__(self):
        return quit_msg
    def __call__(self):
        raise RuntimeError(quit_msg)

_simple_types = (
    bool, int, float, complex, type(None), slice,
    int, # 2to3 replaces long with int, so this is fine
    )
_string_types = (
    bytes if py3k else str, # 2to3 can't replace str with bytes
    str, # 2to3 replaces unicode with str, so this is fine
    )
def is_key_reprable(obj, max_depth=2):
    """
    Check whether an object (which is a dict key) simple enough
    to be used in the completion list.
    This checks that it's of simple types, and has some arbitrary limits.
    """
    if type(obj) in _simple_types:
        return True
    elif type(obj) in _string_types:
        return len(obj) < 1000
    elif type(obj) in (tuple, frozenset):
        if max_depth <= 0 or len(obj) > 5:
            return False
        else:
            return all(is_key_reprable(x, max_depth-1) for x in obj)
    else:
        return False

# SIGINT masking

def can_mask_sigint():
    return sys.platform in ('linux2', 'win32')

def unmask_sigint():
    # We want to mask ctrl-c events when not running user code, to allow
    # the main process to send a SIGINT anytime, in order to allow it to
    # break GUI code executed when idle.
    # If we can't mask ctrl-c events, the main process will only send
    # SIGINT when executing user code (it uses the get_subprocess_info
    # method to know that.)
    # Also, on win32 we may get unwanted ctrl-c events if the user is
    # running a subprocess. Since all processes in the same "console group"
    # get the event, the subprocess may get it before us and exit. Then we 
    # get out of the try-except block, and only later we get the ctrl-c.
    if sys.platform == 'linux2':
        signal.signal(signal.SIGINT, signal.default_int_handler)
    elif sys.platform == 'win32':
        windll.kernel32.SetConsoleCtrlHandler(None, False) #@UndefinedVariable
    else:
        pass

def mask_sigint():
    if sys.platform == 'linux2':
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    elif sys.platform == 'win32':
        windll.kernel32.SetConsoleCtrlHandler(None, True) #@UndefinedVariable
    else:
        pass

class Subprocess(object):
    def __init__(self, port):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect(('localhost', port))

        # Mask SIGINT/Ctrl-C
        mask_sigint()
        
        # Make sys.displayhook change self.last_res
        self.last_res = None
        sys.displayhook = self.displayhook

        # Trick things like pdb into thinking that the namespace we create is
        # the main module
        mainmodule = types.ModuleType('__main__')
        sys.modules['__main__'] = mainmodule
        self.locs = mainmodule.__dict__
        
        # Add '' to sys.path, to be like the regular Python interpreter
        sys.path.insert(0, '')
        
        # Set sys.argv to [''], to be like the regular Python interpreter
        # (Otherwise multiprocessing on win32 starts running subp_main.py)
        sys.argv = ['']

        # Adjust exit and quit objects
        builtins.exit = builtins.quit = Quit()
        
        self.gui_handlers = [GtkHandler(), Qt4Handler(), TkHandler()]
        self.idle_paused = False

        self.gid = 0
        self.flags = 0
        
        # Config
        self.is_pprint = False
        self.is_matplotlib_ia_switch = False
        self.is_matplotlib_ia_warn = False
        self.reshist_size = 0
        
        # Did we already handle matplotlib in non-interactive mode?
        self.matplotlib_ia_handled = False
        
        # The result history index of the next value to enter the history
        self.reshist_counter = 0

        # Run endless loop
        self.loop()

    def loop(self):
        while True:
            if not self.idle_paused:
                self.handle_gui_events(self.sock)
            funcname, args = recv_object(self.sock)
            if funcname in rpc_funcs:
                func = getattr(self, funcname)
                try:
                    r = func(*args)
                    if isinstance(r, types.GeneratorType):
                        for obj in r:
                            send_object(self.sock, obj)
                    else:
                        send_object(self.sock, r)
                except Exception:
                    # This may help in debugging exceptions.
                    traceback.print_exc()
                    send_object(self.sock, None)
            else:
                # aid in debug
                sys.stderr.write("Unknown command: %s\n" % funcname)
                send_object(self.sock, None)

    def displayhook(self, res):
        if res is not None:
            self.last_res = res

    def handle_gui_events(self, sock):
        """
        Handle GUI events until there's something to read from sock.
        If there's no graphic toolkit, just return.
        """
        sock.setblocking(False)
        try:
            while not select([sock], [], [], 0)[0]:
                executed = False
                for handler in self.gui_handlers:
                    cur_executed = handler.handle_events(GUI_SLEEP)
                    executed = executed or cur_executed
                if not executed:
                    break
        finally:
            sock.setblocking(True)

    @rpc_func
    def is_incomplete(self, source):
        """
        Get a string of code. Return True if it's incomplete and False
        otherwise (if it's complete or if there's a syntax error.)
        This is used to run single-line statements without need for
        ctrl+enter.
        """
        try:
            r = codeop.compile_command(source)
        except (SyntaxError, OverflowError, ValueError):
            return False
        return (r is None)
    
    @staticmethod
    def update_features(cur_flags, co_flags):
        """
        Get an int with current __future__ flags, return it updated with
        co_flags from a code object.
        """
        for feature in _features:
            if co_flags & feature.compiler_flag:
                cur_flags |= feature.compiler_flag
        return cur_flags
    
    def compile_ast(self, source):
        """
        Compile source into a list of code objects, updating linecache, self.gid
        and self.flags.
        Return True, codeob on success.
        Return False, reason on syntax error.
        This version uses the ast module available in Python 2.6 and Jython 2.5.
        This version always returns a list with one item.
        """
        filename = '<pyshell#%d>' % self.gid
        try:
            a = compile(source, filename, 'exec',
                        ast.PyCF_ONLY_AST | self.flags)
            b = ast.Interactive(a.body)
            codeob = compile(b, filename, 'single', self.flags)
        except SyntaxError as e:
            # Sometimes lineno or offset are not defined. Zero them in that case.
            lineno = e.lineno if e.lineno is not None else 1
            offset = e.offset if e.offset is not None else 1
            return False, (str(e.msg), lineno-1, offset-1)
        except ValueError as e:
            # Compiling "\x%" raises a ValueError
            return False, (str(e), 0, 0)
            
        # Update gid, linecache, flags
        self.gid += 1
        lines = [x+'\n' for x in source.split("\n")]
        linecache.cache[filename] = len(source)+1, None, lines, filename
        self.flags = self.update_features(self.flags, codeob.co_flags)
        
        return True, [codeob]
    
    def compile_no_ast(self, source):
        """
        This function does the same thing as compile_ast, but it works without
        the ast module.
        """
        split_source = split_to_singles(source)
        # This added newline is because sometimes the CommandCompiler wants
        # more if there isn't a newline at the end
        split_source[-1] += '\n'
        line_count = 0
        # Compile to check for syntax errors
        cur_flags = self.flags
        for src in split_source:
            try:
                c = compile(src, '<pyshell>', 'single', cur_flags)
            except SyntaxError as e:
                # Sometimes lineno or offset are not defined. Zero them in that
                # case.
                lineno = e.lineno if e.lineno is not None else 1
                offset = e.offset if e.offset is not None else 1
                msg = unicodify(e.msg)
                return False, (msg, lineno-1+line_count, offset-1)
            except ValueError as e:
                # Compiling "\x%" raises a ValueError
                return False, (str(e), 0, 0)
            else:
                if c is None:
                    return False, None
                    return
                else:
                    line_count += src.count('\n')
                    cur_flags = self.update_features(cur_flags, c.co_flags)

        # If compilation was successful...
        codeobs = []
        for src in split_source:
            # We compile again, so as not to put into linecache code
            # which had no effect
            filename = '<pyshell#%d>' % self.gid
            self.gid += 1
            lines = [x+'\n' for x in src.split("\n")]
            linecache.cache[filename] = len(src)+1, None, lines, filename
            codeob = compile(src, filename, 'single', self.flags)
            self.flags = self.update_features(self.flags, codeob.co_flags)
            codeobs.append(codeob)
        
        return True, codeobs
    
    @staticmethod
    def safe_pformat(obj):
        """
        Use pprint to format an object.
        In case of an exception, warn and use regular repr instead.
        """
        try:
            return str(pprint.pformat(obj))
        except:
            from warnings import warn
            warn('pprint raised an exception, using repr instead. '
                 'To reproduce, run: "from pprint import pprint; pprint(_)"')
            return str(repr(obj))
    
    @rpc_func
    def execute(self, source):
        """
        Get the source code to execute (a unicode string).
        Compile it. If there was a syntax error, return
        (False, (msg, line, col)).
        If compilation was successful, return (True, None), then run the code
        and then send (is_success, res_no, res_str, exception_string, rem_stdin).
        is_success - True if there was no exception.
        res_no - number of the result in the history count, or None if there
                 was no result or there's no history.
        res_str - a string representation of the result.
        exception_string - description of the exception, or None if is_success.
        rem_stdin - data that was sent into stdin and wasn't consumed.
        """
        # pause_idle was called before execute, disable it.
        self.idle_paused = False
        
        if ast:
            success, r = self.compile_ast(source)
        else:
            success, r = self.compile_no_ast(source)
        if not success:
            yield False, r
            return
        else:
            yield True, None
        codeobs = r
            
        self.last_res = None
        try:
            unmask_sigint()
            try:
                # Execute
                for codeob in codeobs:
                    exec(codeob, self.locs)
                # Work around http://bugs.python.org/issue8213 - stdout buffered
                # in Python 3.
                sys.stdout.flush()
                sys.stderr.flush()
                # Convert the result to a string. This is here because exceptions
                # may be raised here.
                if self.last_res is not None:
                    if self.is_pprint:
                        res_str = self.safe_pformat(self.last_res)
                    else:
                        res_str = str(repr(self.last_res))
                    if len(res_str) > MAX_RES_STR_LEN:
                        res_str = (res_str[:MAX_RES_STR_LEN]
                                   +'\n[%d chars truncated]' % (
                                        len(res_str)-MAX_RES_STR_LEN))
                else:
                    res_str = None
            finally:
                mask_sigint()
        except:
            sys.stdout.flush()
            excinfo = sys.exc_info()
            sys.last_type, sys.last_value, sys.last_traceback = excinfo
            exception_string = trunc_traceback(excinfo, __file__)
            is_success = False
            res_no = None
            res_str = None
        else:
            is_success = True
            exception_string = None
            if self.last_res is not None:
                res_no = self.store_in_reshist(self.last_res)
            else:
                res_no = None
        # Discard the reference to the result
        self.last_res = None
            
        # Send back any data left on stdin.
        rem_stdin = []
        if sys.platform == 'linux2':
            while select([sys.stdin], [], [], 0)[0]:
                r = os.read(sys.stdin.fileno(), 8192)
                if not r:
                    # File may be in error state
                    break
                rem_stdin.append(unicodify(r))
        elif sys.platform == 'win32':
            fd = sys.stdin.fileno()
            handle = get_osfhandle(fd)
            avail = c_ulong(0)
            PeekNamedPipe(handle, None, 0, None, byref(avail), None)
            nAvail = avail.value
            if nAvail > 0:
                rem_stdin.append(os.read(fd, nAvail))
        else:
            # I don't know how to do this in Jython.
            pass

        rem_stdin = ''.join(rem_stdin)
        
        # Check if matplotlib in non-interactive mode was imported
        self.check_matplotlib_ia()

        yield is_success, res_no, res_str, exception_string, rem_stdin

    @rpc_func
    def pause_idle(self):
        """
        before 'execute' is called, 'pause_idle' is called to check if we don't
        do any idle jobs right now. It is followed by either 'execute' or
        'resume_idle'.
        """
        self.idle_paused = True
    
    @rpc_func
    def resume_idle(self):
        self.idle_paused = False
    
    @rpc_func
    def set_pprint(self, is_pprint):
        self.is_pprint = is_pprint
    
    @rpc_func
    def set_matplotlib_ia(self, is_switch, is_warn):
        self.is_matplotlib_ia_switch = is_switch
        self.is_matplotlib_ia_warn = is_warn

    @rpc_func
    def set_reshist_size(self, new_reshist_size):
        if new_reshist_size < self.reshist_size:
            for i in range(self.reshist_counter-self.reshist_size,
                           self.reshist_counter-new_reshist_size):
                self.locs.pop('_%d' % i, None)
        self.reshist_size = new_reshist_size
    
    @rpc_func
    def clear_reshist(self):
        for i in range(self.reshist_counter-self.reshist_size, self.reshist_counter):
            self.locs.pop('_%d' % i, None)

    def store_in_reshist(self, res):
        """
        Get a result value to store in the result history.
        Store it, and return the result's index.
        If the result isn't stored, return None.
        """
        if res is None:
            return None
            
        if '__' in self.locs:
            self.locs['___'] = self.locs['__']
        if '_' in self.locs:
            self.locs['__'] = self.locs['_']
        self.locs['_'] = res
        
        if self.reshist_size == 0:
            return None
        res_index = self.reshist_counter
        self.locs['_%d' % res_index] = res
        del_index = self.reshist_counter - self.reshist_size
        if del_index >= 0:
            self.locs.pop('_%d' % del_index, None)
        self.reshist_counter += 1
        return res_index
    
    @staticmethod
    def split_list(L, public_set):
        """
        split L into two lists: public and private, according to public_set,
        which should be a set of names or None. If it's None, split according
        to whether the first char is '_'.
        """
        public = []
        private = []
        if public_set is not None:
            for x in L:
                if x in public_set:
                    public.append(x)
                else:
                    private.append(x)
        else:
            for x in L:
                if not x.startswith('_'):
                    public.append(x)
                else:
                    private.append(x)
        return public, private

    @rpc_func
    def complete_attributes(self, expr):
        """
        Evaluate expr in the namespace, and return its attributes as two
        sorted lists - public and private.
        public - completions that are thought to be relevant.
        private - completions that are not so.
        """
        try:
            entity = eval(expr, self.locs)
            ids = dir(entity)
            ids = list(map(unicodify, ids))
            ids.sort()
            if hasattr(entity, '__all__'):
                all_set = set(entity.__all__)
            else:
                all_set = None
            public, private = self.split_list(ids, all_set)
        except Exception:
            public = private = []

        return public, private

    @rpc_func
    def complete_firstlevels(self):
        """
        Get (public, private) names (globals, builtins, keywords).
        """
        namespace = self.locs.copy()
        namespace.update(builtins.__dict__)
        ids = eval("dir()", namespace) + keyword.kwlist
        ids = list(map(unicodify, ids))
        ids.sort()
        if '__all__' in namespace:
            all_set = set(namespace['__all__'])
        else:
            all_set = None
        public, private = self.split_list(ids, all_set)
        
        return public, private
        
    @rpc_func
    def get_func_args(self, expr):
        """Return the argument names of the function (a list of strings)"""
        try:
            obj = eval(expr, self.locs)
        except Exception:
            return None
        try:
            if not py3k:
                args = inspect.getargspec(obj)[0]
            else:
                args = inspect.getfullargspec(obj).args #@UndefinedVariable
        except TypeError:
            return None
        # There may be nested args, so we filter them
        return [unicodify(s) for s in args
                if isinstance(s, str)]

    @rpc_func
    def complete_dict_keys(self, expr):
        """
        Return the reprs of the dict's keys (a list of strings)
        Returns only those which have a simple-enough repr.
        """
        try:
            obj = eval(expr, self.locs)
        except Exception:
            return None
        if not isinstance(obj, dict) or len(obj) > 1000:
            return None
        return sorted(unicodify(repr(x)) for x in obj if is_key_reprable(x))
        

    @rpc_func
    def find_modules(self, package):
        if package:
            package = package.split('.')
        else:
            package = []
        return [unicodify(s) for s in find_modules(package)]
    
    @rpc_func
    def get_module_members(self, mod_name):
        try:
            mod = sys.modules[mod_name]
        except KeyError:
            return None
        if hasattr(mod, '__all__'):
            all_set = set(mod.__all__)
        else:
            all_set = None
        ids = [unicodify(x) for x in mod.__dict__.keys()]
        return self.split_list(ids, all_set)
    
    @rpc_func
    def complete_filenames(self, str_prefix, text, str_char, add_quote):
        is_raw = 'r' in str_prefix.lower()
        is_unicode = 'u' in str_prefix.lower()
        try:
            # We add a space because a backslash can't be the last
            # char of a raw string literal
            comp_what = eval(str_prefix
                             + text
                             + ' '
                             + str_char)[:-1]
        except SyntaxError:
            return
        if comp_what == '':
            comp_what = '.'
        if comp_what.startswith('//'):
            # This may be an XPath expression. Calling listdir on win32 will
            # interpret this as UNC and search the network, which may take
            # a long time or even stall. You can still use r'\\...' if you
            # want completion from UNC paths.
            return

        try:
            dirlist = os.listdir(comp_what)
        except OSError:
            return
        if case_insen_filenames:
            dirlist.sort(key=lambda s: s.lower())
        else:
            dirlist.sort()
        public = []
        private = []
        for name in dirlist:
            orig_name = name
            if not py3k:
                if is_unicode and isinstance(name, str):
                    # A filename which can't be unicode
                    continue
                if not is_unicode:
                    # We need a unicode string as the code. From what I see,
                    # Python evaluates unicode characters in byte strings as utf-8.
                    try:
                        name = name.decode('utf8')
                    except UnicodeDecodeError:
                        continue
            # skip troublesome names
            try:
                rename = eval(str_prefix + name + str_char)
            except (SyntaxError, UnicodeDecodeError):
                continue
            if rename != orig_name:
                continue

            is_dir = os.path.isdir(os.path.join(comp_what, orig_name))

            if not is_dir:
                if add_quote:
                    name += str_char
            else:
                if '/' in text or os.path.sep == '/':
                    # Prefer forward slash
                    name += '/'
                else:
                    if not is_raw:
                        name += '\\\\'
                    else:
                        name += '\\'

            if name.startswith('.'):
                private.append(name)
            else:
                public.append(name)
        
        return public, private, case_insen_filenames
    
    @staticmethod
    def get_welcome():
        if 'IronPython' in sys.version:
            first_line = sys.version[sys.version.find('(')+1:sys.version.rfind(')')]
        else:
            if sys.platform.startswith('java'):
                name = 'Jython'
            else:
                name = 'Python'
            first_line = '%s %s on %s' % (name, sys.version, sys.platform)
        return (first_line+'\n'
                +'Type "copyright", "credits" or "license()" for more information.\n')

    @rpc_func
    def get_subprocess_info(self):
        return (self.get_welcome(), can_mask_sigint())
        
    @classmethod
    def _find_constructor(cls, class_ob):
        # Given a class object, return a function object used for the
        # constructor (ie, __init__() ) or None if we can't find one.
        try:
            return class_ob.__init__.__func__
        except AttributeError:
            for base in class_ob.__bases__:
                rc = cls._find_constructor(base)
                if rc is not None: return rc
        return None

    @rpc_func
    def get_func_doc(self, expr):
        """Get a string describing the arguments for the given object"""
        try:
            obj = eval(expr, self.locs)
        except Exception:
            return None
        if isinstance(obj, (types.BuiltinFunctionType,
                            types.BuiltinMethodType)):
            # These don't have source code, and using pydoc will only
            # add something like "execfile(...)" before the doc.
            doc = inspect.getdoc(obj)
            if doc is None:
                return None
            return unicodify(doc)
        
        # Check if obj.__doc__ is not in the code (was added after definition).
        # If so, return pydoc's documentation.
        # This test is CPython-specific. Another approach would be to look for
        # the string in the source code.
        co_consts = getattr(getattr(obj, 'func_code', None), 'co_consts', None)
        __doc__ = getattr(obj, '__doc__', None)
        if co_consts is not None and __doc__ is not None:
            if __doc__ not in co_consts:
                # Return pydoc's documentation
                return unicodify(textdoc.document(obj).strip())
        
        try:
            source = inspect.getsource(obj)
        except (TypeError, IOError):
            # If can't get the source, return pydoc's documentation
            return unicodify(textdoc.document(obj).strip())
        else:
            # If we can get the source, return it.
            
            # cleandoc removes extra indentation.
            # We add a newline because it ignores indentation of first line...
            # The next line is for Python 2.5 compatibility.
            cleandoc = getattr(inspect, 'cleandoc', lambda s: s)
            return unicodify(cleandoc('\n'+source))
    
    @rpc_func
    def is_callable_only(self, what):
        """
        Checks whether an object is callable, and doesn't expect operators
        (so there's no point in typing space after its name, unless to add
        parens).
        Also checks whether obj.__expects_str__ is True, which means that
        the expected argument is a string so quotes will be added.
        Returns (is_callable_only, expects_str)
        """
        try:
            obj = eval(what, self.locs)
        except Exception:
            return False, False
        typ = type(obj)
        
        expects_str = bool(getattr(obj, '__expects_str__', False))
        
        # Check cache
        try:
            return is_callable_cache[id(typ)], expects_str
        except KeyError:
            pass
        
        r = (isinstance(obj, collections.Callable)
             and not any(hasattr(obj, att) for att in operator_methods))
        
        is_callable_cache[id(typ)] = r
        return r, expects_str
    
    def check_matplotlib_ia(self):
        """Check if matplotlib is in non-interactive mode, and handle it."""
        if not self.is_matplotlib_ia_warn and not self.is_matplotlib_ia_switch:
            return
        if self.matplotlib_ia_handled:
            return
        if 'matplotlib' not in sys.modules:
            return
        self.matplotlib_ia_handled = True
        # From here we do this only once.
        matplotlib = sys.modules['matplotlib']
        if not hasattr(matplotlib, 'is_interactive'):
            return
        if matplotlib.is_interactive():
            return
        if self.is_matplotlib_ia_switch:
            if not hasattr(matplotlib, 'interactive'):
                return
            matplotlib.interactive(True)
        else:
            sys.stderr.write(
                "Warning: matplotlib in non-interactive mode detected.\n"
                "This means that plots will appear only after you run show().\n"
                "Use Edit->Preferences->Shell to automatically switch to interactive mode \n"
                "or to suppress this warning.\n")


# Handle GUI events

class GuiHandler(object):
    def handle_events(self, delay):
        """
        This method gets the time in which to process GUI events, in seconds.
        If the GUI toolkit is loaded, run it for the specified delay and return
        True.
        If it isn't loaded, return False immediately.
        """
        raise NotImplementedError("Abstract method")

@contextmanager
def user_code():
    """
    Run user code, unmasking SIGINT, and catching exceptions.
    """
    try:
        unmask_sigint()
        yield
    except:
        sys.excepthook(*sys.exc_info())
    finally:
        mask_sigint()


class GtkHandler(GuiHandler):
    def __init__(self):
        self.gtk = None
        self.timeout_add = None

    def handle_events(self, delay):
        if self.gtk is None:
            if 'gtk' in sys.modules:
                self.gtk = sys.modules['gtk']
                try:
                    from glib import timeout_add
                except ImportError:
                    from gobject import timeout_add
                self.timeout_add = timeout_add
            else:
                return False
        self.timeout_add(int(delay * 1000), self.gtk_main_quit)
        with user_code():
            self.gtk.main()
            
        return True

    def gtk_main_quit(self):
        self.gtk.main_quit()
        # Don't call me again
        return False

class Qt4Handler(GuiHandler):
    def __init__(self):
        self.QtCore = None
        self.app = None

    def handle_events(self, delay):
        if self.QtCore is None:
            if 'PyQt4' in sys.modules:
                self.QtCore = sys.modules['PyQt4'].QtCore
            else:
                return False
        QtCore = self.QtCore

        if self.app is None:
            app = QtCore.QCoreApplication.instance()
            if app:
                self.app = app
            else:
                return False

        timer = QtCore.QTimer()
        QtCore.QObject.connect(timer, QtCore.SIGNAL('timeout()'),
                               self.qt4_quit_if_no_modal)
        timer.start(delay*1000)
        with user_code():
            self.app.exec_()
        timer.stop()
        QtCore.QObject.disconnect(timer, QtCore.SIGNAL('timeout()'),
                                  self.qt4_quit_if_no_modal)
        return True

    def qt4_quit_if_no_modal(self):
        app = self.app
        if app.__class__.__name__ != 'QApplication' or \
           app.activeModalWidget() is None:
            app.quit()

class TkHandler(GuiHandler):
    def __init__(self):
        self.Tkinter = None

    def handle_events(self, delay):
        # TODO: It's pretty silly to handle all events and then just wait.
        # But I haven't found a better way - if you find one, tell me!
        if self.Tkinter is None:
            if 'Tkinter' in sys.modules:
                self.Tkinter = sys.modules['Tkinter']
            else:
                return False
        Tkinter = self.Tkinter

        # Handling Tk events is done only if there is an active tkapp object.
        # It is created by Tkinter.Tk.__init__, which sets
        # Tkinter._default_root to itself, when Tkinter._support_default_root
        # is True (the default). Here we check whether Tkinter._default_root
        # is something before we handle Tk events.
        if Tkinter._default_root:
            _tkinter = Tkinter._tkinter
            with user_code():
                while _tkinter.dooneevent(_tkinter.DONT_WAIT):
                    pass
        time.sleep(delay)
        return True

def main(port):
    _subp = Subprocess(port)



# This file is based on brine.py from RPyC.
# See http://rpyc.wikidot.com/
# and http://sebulbasvn.googlecode.com/svn/tags/rpyc/3.0.6/core/brine.py
# Modified by Noam Yorav-Raphael for DreamPie use.

# Copyright (c) 2005-2009
# Tomer Filiba (tomerfiliba@gmail.com)
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

"""
brine - a simple, fast and secure object serializer,
optimized for small integers [-48..160), suitable for Python 2/3k communication.
the following types are supported: int (in the unsigned long range), bool,
unicode (In Py2) / str (In Py3), float, slice, complex, tuple(of simple types),
list(of simple types), frozenset(of simple types)
as well as the following singletons: None, NotImplemented, Ellipsis
"""
import sys
py3k = (sys.version_info[0] == 3)
if not py3k:
    from cStringIO import StringIO
else:
    from io import BytesIO
from struct import Struct

if not py3k:
    def b(n):
        return chr(n)
    empty_bytes = ''
else:
    def b(n):
        return bytes([n])
    empty_bytes = bytes()

# singletons
TAG_NONE = b(0x00)
TAG_EMPTY_STR = b(0x01)
TAG_EMPTY_TUPLE = b(0x02)
TAG_TRUE = b(0x03)
TAG_FALSE = b(0x04)
TAG_NOT_IMPLEMENTED = b(0x05)
TAG_ELLIPSIS = b(0x06)
# types
#TAG_UNICODE = b(0x08) # Removed - STR is unicode.
#TAG_LONG = b(0x09) # Removed
TAG_STR1 = b(0x0a)
TAG_STR2 = b(0x0b)
TAG_STR3 = b(0x0c)
TAG_STR4 = b(0x0d)
TAG_STR_L1 = b(0x0e)
TAG_STR_L4 = b(0x0f)
TAG_TUP1 = b(0x10)
TAG_TUP2 = b(0x11)
TAG_TUP3 = b(0x12)
TAG_TUP4 = b(0x13)
TAG_TUP_L1 = b(0x14)
TAG_TUP_L4 = b(0x15)
TAG_INT_L1 = b(0x16)
TAG_INT_L4 = b(0x17)
TAG_FLOAT = b(0x18)
TAG_SLICE = b(0x19)
TAG_FSET = b(0x1a)
TAG_COMPLEX = b(0x1b)

# List
TAG_EMPTY_LIST = b(0x1c)
TAG_LIST1 = b(0x1d)
TAG_LIST_L1 = b(0x1e)
TAG_LIST_L4 = b(0x1f)

IMM_INTS = dict((i, b(i + 0x50)) for i in range(-0x30, 0xa0))

I1 = Struct("!B")
I4 = Struct("!L")
F8 = Struct("!d")
C16 = Struct("!dd")

_dump_registry = {}
_load_registry = {}
IMM_INTS_LOADER = dict((v, k) for k, v in IMM_INTS.iteritems())

def register(coll, key):
    def deco(func):
        coll[key] = func
        return func
    return deco

#===============================================================================
# dumping
#===============================================================================
@register(_dump_registry, type(None))
def _dump_none(_obj, stream):
    stream.append(TAG_NONE)

@register(_dump_registry, type(NotImplemented))
def _dump_notimplemeted(_obj, stream):
    stream.append(TAG_NOT_IMPLEMENTED)

@register(_dump_registry, type(Ellipsis))
def _dump_ellipsis(_obj, stream):
    stream.append(TAG_ELLIPSIS)

@register(_dump_registry, bool)
def _dump_bool(obj, stream):
    if obj:
        stream.append(TAG_TRUE)
    else:
        stream.append(TAG_FALSE)

@register(_dump_registry, slice)
def _dump_slice(obj, stream):
    stream.append(TAG_SLICE)
    _dump((obj.start, obj.stop, obj.step), stream)

@register(_dump_registry, frozenset)
def _dump_frozenset(obj, stream):
    stream.append(TAG_FSET)
    _dump(tuple(obj), stream)

@register(_dump_registry, int)
def _dump_int(obj, stream):
    if obj in IMM_INTS:
        stream.append(IMM_INTS[obj])
    else:
        obj = str(obj)
        l = len(obj)
        if l < 256:
            stream.append(TAG_INT_L1 + I1.pack(l) + obj)
        else:
            stream.append(TAG_INT_L4 + I4.pack(l) + obj)

#@register(_dump_registry, long)
#def _dump_long(obj, stream):
#    stream.append(TAG_LONG)
#    _dump_int(obj, stream)

@register(_dump_registry, unicode)
def _dump_str(obj, stream):
    obj = obj.encode('utf8')
    l = len(obj)
    if l == 0:
        stream.append(TAG_EMPTY_STR)
    elif l == 1:
        stream.append(TAG_STR1 + obj)
    elif l == 2:
        stream.append(TAG_STR2 + obj)
    elif l == 3:
        stream.append(TAG_STR3 + obj)
    elif l == 4:
        stream.append(TAG_STR4 + obj)
    elif l < 256:
        stream.append(TAG_STR_L1 + I1.pack(l) + obj)
    else:
        stream.append(TAG_STR_L4 + I4.pack(l) + obj)

@register(_dump_registry, float)
def _dump_float(obj, stream):
    stream.append(TAG_FLOAT + F8.pack(obj))

@register(_dump_registry, complex)
def _dump_complex(obj, stream):
    stream.append(TAG_COMPLEX + C16.pack(obj.real, obj.imag))

#@register(_dump_registry, unicode)
#def _dump_unicode(obj, stream):
#    stream.append(TAG_UNICODE)
#    _dump_str(obj.encode("utf8"), stream)

@register(_dump_registry, tuple)
def _dump_tuple(obj, stream):
    l = len(obj)
    if l == 0:
        stream.append(TAG_EMPTY_TUPLE)
    elif l == 1:
        stream.append(TAG_TUP1)
    elif l == 2:
        stream.append(TAG_TUP2)
    elif l == 3:
        stream.append(TAG_TUP3)
    elif l == 4:
        stream.append(TAG_TUP4)
    elif l < 256:
        stream.append(TAG_TUP_L1 + I1.pack(l))
    else:
        stream.append(TAG_TUP_L4 + I4.pack(l))
    for item in obj:
        _dump(item, stream)

@register(_dump_registry, list)
def _dump_list(obj, stream):
    l = len(obj)
    if l == 0:
        stream.append(TAG_EMPTY_LIST)
    elif l == 1:
        stream.append(TAG_LIST1)
    elif l < 256:
        stream.append(TAG_LIST_L1 + I1.pack(l))
    else:
        stream.append(TAG_LIST_L4 + I4.pack(l))
    for item in obj:
        _dump(item, stream)

def _undumpable(obj, stream):
    raise TypeError("cannot dump %r" % (obj,))

def _dump(obj, stream):
    _dump_registry.get(type(obj), _undumpable)(obj, stream)

#===============================================================================
# loading
#===============================================================================
@register(_load_registry, TAG_NONE)
def _load_none(_stream):
    return None
@register(_load_registry, TAG_NOT_IMPLEMENTED)
def _load_nonimp(_stream):
    return NotImplemented
@register(_load_registry, TAG_ELLIPSIS)
def _load_elipsis(_stream):
    return Ellipsis
@register(_load_registry, TAG_TRUE)
def _load_true(_stream):
    return True
@register(_load_registry, TAG_FALSE)
def _load_false(_stream):
    return False
@register(_load_registry, TAG_EMPTY_TUPLE)
def _load_empty_tuple(_stream):
    return ()
@register(_load_registry, TAG_EMPTY_LIST)
def _load_empty_list(_stream):
    return []
@register(_load_registry, TAG_EMPTY_STR)
def _load_empty_str(_stream):
    return u""
#@register(_load_registry, TAG_UNICODE)
#def _load_unicode(stream):
#    obj = _load(stream)
#    return obj.decode("utf-8")
#@register(_load_registry, TAG_LONG)
#def _load_long(stream):
#    obj = _load(stream)
#    return long(obj)

@register(_load_registry, TAG_FLOAT)
def _load_float(stream):
    return F8.unpack(stream.read(8))[0]
@register(_load_registry, TAG_COMPLEX)
def _load_complex(stream):
    real, imag = C16.unpack(stream.read(16))
    return complex(real, imag)

@register(_load_registry, TAG_STR1)
def _load_str1(stream):
    return stream.read(1).decode('utf8')
@register(_load_registry, TAG_STR2)
def _load_str2(stream):
    return stream.read(2).decode('utf8')
@register(_load_registry, TAG_STR3)
def _load_str3(stream):
    return stream.read(3).decode('utf8')
@register(_load_registry, TAG_STR4)
def _load_str4(stream):
    return stream.read(4).decode('utf8')
@register(_load_registry, TAG_STR_L1)
def _load_str_l1(stream):
    l, = I1.unpack(stream.read(1))
    return stream.read(l).decode('utf8')
@register(_load_registry, TAG_STR_L4)
def _load_str_l4(stream):
    l, = I4.unpack(stream.read(4))
    return stream.read(l).decode('utf8')

@register(_load_registry, TAG_TUP1)
def _load_tup1(stream):
    return (_load(stream),)
@register(_load_registry, TAG_TUP2)
def _load_tup2(stream):
    return (_load(stream), _load(stream))
@register(_load_registry, TAG_TUP3)
def _load_tup3(stream):
    return (_load(stream), _load(stream), _load(stream))
@register(_load_registry, TAG_TUP4)
def _load_tup4(stream):
    return (_load(stream), _load(stream), _load(stream), _load(stream))
@register(_load_registry, TAG_TUP_L1)
def _load_tup_l1(stream):
    l, = I1.unpack(stream.read(1))
    return tuple(_load(stream) for i in range(l))
@register(_load_registry, TAG_TUP_L4)
def _load_tup_l4(stream):
    l, = I4.unpack(stream.read(4))
    return tuple(_load(stream) for i in xrange(l))

@register(_load_registry, TAG_LIST1)
def _load_list1(stream):
    return [_load(stream)]
@register(_load_registry, TAG_LIST_L1)
def _load_list_l1(stream):
    l, = I1.unpack(stream.read(1))
    return list(_load(stream) for i in range(l))
@register(_load_registry, TAG_LIST_L4)
def _load_list_l4(stream):
    l, = I4.unpack(stream.read(4))
    return list(_load(stream) for i in xrange(l))


@register(_load_registry, TAG_SLICE)
def _load_slice(stream):
    start, stop, step = _load(stream)
    return slice(start, stop, step)
@register(_load_registry, TAG_FSET)
def _load_frozenset(stream):
    return frozenset(_load(stream))

@register(_load_registry, TAG_INT_L1)
def _load_int_l1(stream):
    l, = I1.unpack(stream.read(1))
    return int(stream.read(l))
@register(_load_registry, TAG_INT_L4)
def _load_int_l4(stream):
    l, = I4.unpack(stream.read(4))
    return int(stream.read(l))

def _load(stream):
    tag = stream.read(1)
    if tag in IMM_INTS_LOADER:
        return IMM_INTS_LOADER[tag]
    return _load_registry.get(tag)(stream)

#===============================================================================
# API
#===============================================================================
def dump(obj):
    """dumps the given object to a byte-string representation"""
    stream = []
    _dump(obj, stream)
    return empty_bytes.join(stream)

def load(data):
    """loads the given byte-string representation to an object"""
    if not py3k:
        stream = StringIO(data)
    else:
        stream = BytesIO(data)
    return _load(stream)


simple_types = frozenset([type(None), int, long, bool, str, float, unicode, 
    slice, complex, type(NotImplemented), type(Ellipsis)])
def dumpable(obj):
    """indicates whether the object is dumpable by brine"""
    if type(obj) in simple_types:
        return True
    if type(obj) in (tuple, list, frozenset):
        return all(dumpable(item) for item in obj)
    return False


if __name__ == "__main__":
    x = (u"he", 7, u"llo", 8, (), 900, None, True, Ellipsis, 18.2, 18.2j + 13, 
        slice(1,2,3), frozenset([5,6,7]), [8,9,10], NotImplemented)
    assert dumpable(x)
    y = dump(x)
    z = load(y)
    assert x == z
    











# This file was automatically generated from a file in the source DreamPie dir.
# DO NOT EDIT IT, as your changes will be gone when the file is created again.

# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['split_to_singles']

import tokenize
import itertools

class ReadLiner(object):
    """
    Perform readline over a string.
    After finishing, line_offsets contains the offset in the string for each
    line. Each line, except for the last one, ends with a '\n'. The last line
    doesn't end with a '\n'. So the number of lines is the number of '\n' chars
    in the string plus 1.
    """
    def __init__(self, s):
        self.s = s
        self.line_offsets = [0]
        self.finished = False

    def __call__(self):
        if self.finished:
            return ''
        s = self.s
        line_offsets = self.line_offsets
        next_offset = s.find('\n', line_offsets[-1])
        if next_offset == -1:
            self.finished = True
            return s[line_offsets[-1]:]
        else:
            line_offsets.append(next_offset+1)
            return s[line_offsets[-2]:line_offsets[-1]]

class TeeIter(object):
    """Wrap an iterable to add a tee() method which tees."""
    def __init__(self, iterable):
        self._it = iterable
    
    def __iter__(self):
        return self
    
    def next(self):
        return self._it.next()
    
    def tee(self):
        self._it, r = itertools.tee(self._it)
        return r

def split_to_singles(source):
    """Get a source string, and split it into several strings,
    each one a "single block" which can be compiled in the "single" mode.
    Every string which is not the last one ends with a '\n', so to convert
    a line number of a sub-string to a line number of the big string, add
    the number of '\n' chars in the preceding strings.
    """
    readline = ReadLiner(source)
    first_lines = [0] # Indices, 0-based, of the rows which start a new single.
    cur_indent_level = 0
    had_decorator = False
    
    # What this does is pretty simple: We split on every NEWLINE token which
    # is on indentation level 0 and is not followed by "else", "except" or
    # "finally" (in that case it should be kept with the previous "single").
    # Since we get the tokens one by one, and INDENT and DEDENT tokens come
    # *after* the NEWLINE token, we need a bit of care, so we peek at tokens
    # after the NEWLINE token to decide what to do.
    
    tokens_iter = TeeIter(
        itertools.ifilter(lambda x: x[0] not in (tokenize.COMMENT, tokenize.NL),
                          tokenize.generate_tokens(readline)))
    try:
        for typ, s, (srow, _scol), (_erow, _rcol), line in tokens_iter:
            if typ == tokenize.NEWLINE:
                for typ2, s2, (_srow2, _scol2), (_erow2, _rcol2), _line2 \
                    in tokens_iter.tee():
                    if typ2 == tokenize.INDENT:
                        cur_indent_level += 1
                    elif typ2 == tokenize.DEDENT:
                        cur_indent_level -= 1
                    else:
                        break
                else:
                    raise AssertionError("Should have received an ENDMARKER")
                # Now we have the first token after INDENT/DEDENT ones.
                if (cur_indent_level == 0
                    and (typ2 != tokenize.ENDMARKER
                         and not (typ2 == tokenize.NAME
                                  and s2 in ('else', 'except', 'finally')))):
                    if not had_decorator:
                        first_lines.append(srow)
                    else:
                        had_decorator = False

            elif s == '@' and cur_indent_level == 0:
                # Skip next first-line
                had_decorator = True
                        
                        
    except tokenize.TokenError:
        # EOF in the middle, it's a syntax error anyway.
        pass
        
    line_offsets = readline.line_offsets
    r = []
    for i, line in enumerate(first_lines):
        if i != len(first_lines)-1:
            r.append(source[line_offsets[line]:line_offsets[first_lines[i+1]]])
        else:
            r.append(source[line_offsets[line]:])
    return r

tests = [
"""
a = 3
""","""
a = 3
b = 5
""","""
if 1:
    1
""","""
if 1:
    2
else:
    3
""","""
if 1:
    1
if 1:
    2
else:
    3
# comment
""","""
try:
    1/0
except:
    print 'oops'
""","""
def f():
    a = 3
    def g():
        a = 4
f()
""","""
def f():
    a = 3
    def g():
        a = 4
f()
# comment
""","""
try:
    1
finally:
    2
""","""
a=3
if 1:
# comment
    2
    # comment
# comment
else:
    3
""","""
@dec

def f():
    pass
""","""
if 1:
    pass
    
@dec

def f():
    pass
""","""
class Class:
    @dec
    def method():
        pass

def f():
    pass
"""
]

def test():
    # This should raise a SyntaxError if splitting wasn't right.
    for t in tests:
        singles = split_to_singles(t)
        for s in singles:
            compile(s, "fn", "single")
    print "Test was successful"


#!/usr/bin/env python

import sys
import os
from os.path import join

from distutils.core import setup

from dreampielib import __version__, subp_lib

# What's non-standard about installing DreamPie is:
# * There are automatically generated modules, in dreampielib/data.
#   These are generated whenever the setup.py script is run, so from the
#   distutils point of view these files are regular package data files.
# * py2exe doesn't support package data, so "regular" data_files are used,
#   and the 'data' directory ends up near the executable.
# * when py2exe is used, the version is automatically set in setup.nsi.

subp_lib.build()

if 'py2exe' in sys.argv:
    import py2exe #@UnresolvedImport
else:
    py2exe = None

package_data_files = (['data/dreampie.glade',
                       'data/dreampie.png',
                       'data/subp_main.py'] + 
                      [join('data', libfn, fn)
                       for libfn in subp_lib.lib_fns.values()
                       for fn in subp_lib.files])

if py2exe is not None:
    # Add files normally installed in package_data to data_files
    d = {}
    for fn in package_data_files:
        d.setdefault(os.path.dirname(fn), []).append(join('dreampielib', fn))
    additional_py2exe_data_files = d.items()
    
    # Update setup.nsi
    mydir = os.path.dirname(os.path.abspath(__file__))
    template = open(join(mydir, 'setup.nsi.in')).read()
    prod_ver_num_short = map(int, __version__.split('.'))
    prod_ver_num = prod_ver_num_short + [0] * (4-len(prod_ver_num_short))
    prod_ver = '.'.join(map(str, prod_ver_num))
    setup_nsi = ('# Generate from setup.nsi.in. DO NOT EDIT.\n'
                 + template.replace('{AUTO_VERSION}', __version__)
                           .replace('{AUTO_PRODUCT_VERSION}', prod_ver))
    f = open(join(mydir, 'setup.nsi'), 'w')
    f.write(setup_nsi)
    f.close()
    
    
else:
    additional_py2exe_data_files = []


setup_args = dict(
    name='dreampie',
    version=__version__,
    description="DreamPie - The interactive Python shell you've always dreamed about!",
    author='Noam Yorav-Raphael',
    author_email='noamraph@gmail.com',
    url='http://www.dreampie.org/',
    license='GPL v3+',
    scripts=['dreampie'],
    packages=['dreampielib',
              'dreampielib.common', 'dreampielib.gui', 'dreampielib.subprocess',
              ],
    package_data={'dreampielib': package_data_files},
    data_files=[
                ('share/applications', ['share/applications/dreampie.desktop']),
                ('share/man/man1', ['share/man/man1/dreampie.1']),
                ('share/pixmaps', ['share/pixmaps/dreampie.svg',
                                   'share/pixmaps/dreampie.png']),
               ] + additional_py2exe_data_files,
    zip_safe=False,
    )

if py2exe is not None:
    setup_args.update(dict(
        console=[{'script': 'dreampie.py',
                  'icon_resources': [(1, 'dreampie.ico')]}],
        windows=[{'script': 'create-shortcuts.py',
                  'icon_resources': [(1, 'blank.ico')]}],
        options={'py2exe':
                 {'excludes':['_scproxy', 'glib', 'gobject', 'gtk',
                             'gtk.gdk', 'gtk.glade', 'gtksourceview2',
                             'pango', 'pygtk', 'runtime', 'comtypes.gen',
                             '_ssl', 'doctest', 'pdb', 'unittest', 'difflib',
                             'unicodedata', 'bz2', 'zipfile', 'lib2to3',
                             'dulwich', 'dulwich.repo', 'win32api', 'win32con',
                             'win32pipe', 'Carbon', 'Carbon.Files', 'decimal'],
                  'includes':['fnmatch', 'glob', 'ctypes.util'],
                  
                 }},
    ))

setup(**setup_args)



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Selection']

import gtk

from .tags import COMMAND, PROMPT
from .common import beep, get_text

class Selection(object):
    """
    Handle clipboard events.
    When something is selected, "Copy" should be enabled. When nothing is
    selected, "Interrupt" should be enabled.
    Also, "copy only commands" command.
    """
    def __init__(self, textview, sourceview, sv_changed,
                 on_is_something_selected_changed):
        self.textview = textview
        self.textbuffer = textview.get_buffer()
        self.sourceview = sourceview
        self.sourcebuffer = sourceview.get_buffer()
        sv_changed.append(self.on_sv_changed)
        self.on_is_something_selected_changed = on_is_something_selected_changed
        
        self.is_something_selected = None
        self.textbuffer.connect('mark-set', self.on_mark_set)
        self.mark_set_handler = self.sourcebuffer.connect('mark-set',
                                                          self.on_mark_set)
        self.clipboard = gtk.Clipboard()

    def on_sv_changed(self, new_sv):
        self.sourcebuffer.disconnect(self.mark_set_handler)
        self.sourceview = new_sv
        self.sourcebuffer = new_sv.get_buffer()
        self.mark_set_handler = self.sourcebuffer.connect('mark-set',
                                                          self.on_mark_set)
    
    def on_selection_changed(self, _clipboard, _event):
        is_something_selected = (self.textbuffer.get_has_selection()
                                 or self.sourcebuffer.get_has_selection())
        self.on_is_something_selected_changed(is_something_selected)

    def on_mark_set(self, _widget, _it, _mark):
        is_something_selected = (self.textbuffer.get_has_selection()
                                 or self.sourcebuffer.get_has_selection())
        if self.is_something_selected is None \
           or is_something_selected != self.is_something_selected:
            self.is_something_selected = is_something_selected
            self.on_is_something_selected_changed(is_something_selected)

    def cut(self):
        if self.sourcebuffer.get_has_selection():
            self.sourcebuffer.cut_clipboard(self.clipboard, True)
        else:
            beep()

    def copy(self):
        if self.textbuffer.get_has_selection():
            # Don't copy '\r' chars, which are newlines only used for
            # display
            tb = self.textbuffer
            sel_start, sel_end = tb.get_selection_bounds()
            text = get_text(tb, sel_start, sel_end)
            text = text.replace('\r', '')
            self.clipboard.set_text(text)
        elif self.sourcebuffer.get_has_selection():
            self.sourcebuffer.copy_clipboard(self.clipboard)
        else:
            beep()

    def copy_commands_only(self):
        if self.sourcebuffer.get_has_selection():
            self.sourcebuffer.copy_clipboard(self.clipboard)
            return
        if not self.textbuffer.get_has_selection():
            beep()
            return
        # We need to copy the text which has the COMMAND tag, doesn't have
        # the PROMPT tag, and is selected.
        tb = self.textbuffer
        command = tb.get_tag_table().lookup(COMMAND)
        prompt = tb.get_tag_table().lookup(PROMPT)
        r = []
        it, sel_end = tb.get_selection_bounds()
        reached_end = False
        while not reached_end:
            it2 = it.copy()
            it2.forward_to_tag_toggle(None)
            if it2.compare(sel_end) >= 0:
                it2 = sel_end.copy()
                reached_end = True
            if it.has_tag(command) and not it.has_tag(prompt):
                r.append(get_text(tb, it, it2))
            it = it2
        r = ''.join(r)
        if not r:
            beep()
        else:
            self.clipboard.set_text(r)

    def paste(self):
        if self.sourceview.is_focus():
            self.sourcebuffer.paste_clipboard(self.clipboard, None, True)
        else:
            beep()
